#
# This file is part of usb-protocol.
#
""" Structures describing Communications Device Class descriptors. """

import unittest
from enum import IntEnum

import construct
from   construct  import this, Default

from .. import LanguageIDs
from ..descriptor import \
    DescriptorField, DescriptorNumber, DescriptorFormat, \
    BCDFieldAdapter, DescriptorLength


class CDCDescriptorNumbers(IntEnum):
    CS_INTERFACE  = 0x24
    CS_ENDPOINT   = 0x25


class CDCDescriptorSubtypes(IntEnum):
    """ Numbers of the Communications Class descriptor subtypes. """

    HEADER                                    = 0x00
    CALL_MANAGEMENT_FUNCTIONAL                = 0x01
    ABSTRACT_CONTROL_MANAGEMENT_FUNCTIONAL    = 0x02
    DIRECT_LINE_MANAGEMENT_FUNCTIONAL         = 0x03
    TELEPHONE_RINGER_FUNCTIONAL               = 0x04
    TELEPHONE_CALL_AND_LINE_STATE_FUNCTIONAL  = 0x05
    UNION_FUNCTIONAL_DESCRIPTOR               = 0x06
    COUNTRY_SELECTION_FUNCTIONAL              = 0x07
    TELEPHONE_OPERATIONAL_MODES_FUNCTIONAL    = 0x08
    USB_TERMINAL_FUNCTIONAL                   = 0x09
    NETWORK_CHANNEL_TERMINAL                  = 0x0a
    PROTOCOL_UNIT_FUNCTIONAL                  = 0x0b
    EXTENSION_UNIT_FUNCTIONAL                 = 0x0c
    MULTI_CHANNEL_MANAGEMENT_FUNCTIONAL       = 0x0d
    CAPI_CONTROL_MANAGEMENT_FUNCTIONAL        = 0x0e
    ETHERNET_NETWORKING_FUNCTIONAL            = 0x0f
    ATM_NETWORKING_FUNCTIONAL                 = 0x10
    WIRELESS_HANDSET_CONTROL_MODEL_FUNCTIONAL = 0x11
    MOBILE_DIRECT_LINE_MODEL_FUNCTIONAL       = 0x12
    MDLM_DETAIL_FUNCTIONAL                    = 0x13
    DEVICE_MANAGEMENT_MODEL_FUNCTIONAL        = 0x14
    OBEX_FUNCTIONAL                           = 0x15
    COMMAND_SET_FUCNTIONAL                    = 0x16
    COMMAND_SET_DETAIL_FUNCTIONAL             = 0x17
    TELEPHONE_CONTROL_MODEL_FUNCTIONAL        = 0x18
    OBEX_SERVICE_IDENTIFIER_FUNCTIONAL        = 0x19
    NCM_FUNCTIONAL                            = 0x1A


HeaderDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(5, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(CDCDescriptorNumbers.CS_INTERFACE),
    "bDescriptorSubtype"  / DescriptorNumber(CDCDescriptorSubtypes.HEADER),
    "bcdCDC"              / DescriptorField(description="CDC Version", default=1.1)
)

ACMFunctionalDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(4, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(CDCDescriptorNumbers.CS_INTERFACE),
    "bDescriptorSubtype"  / DescriptorNumber(CDCDescriptorSubtypes.ABSTRACT_CONTROL_MANAGEMENT_FUNCTIONAL),
    "bmCapabilities"      / DescriptorField(description="ACM Capabilities", default=0b0010)
)

UnionFunctionalDescriptor = DescriptorFormat(
    "bLength"                / construct.Const(5, construct.Int8ul),
    "bDescriptorType"        / DescriptorNumber(CDCDescriptorNumbers.CS_INTERFACE),
    "bDescriptorSubtype"     / DescriptorNumber(CDCDescriptorSubtypes.UNION_FUNCTIONAL_DESCRIPTOR),
    "bControlInterface"      / DescriptorField(description="Control Interface Number"),
    "bSubordinateInterface0" / DescriptorField(description="Subordinate Interface Number")
)

CallManagementFunctionalDescriptor = DescriptorFormat(
    "bLength"                / construct.Const(5, construct.Int8ul),
    "bDescriptorType"        / DescriptorNumber(CDCDescriptorNumbers.CS_INTERFACE),
    "bDescriptorSubtype"     / DescriptorNumber(CDCDescriptorSubtypes.CALL_MANAGEMENT_FUNCTIONAL),
    "bmCapabilities"         / DescriptorField(description="Call Management capabilities", default=0),
    "bDataInterface"         / DescriptorField(description="Data Interface Number")
)



#
# This file is part of usb-protocol.
#
""" Helpers for creating construct-related emitters. """

import unittest
import construct


class ConstructEmitter:
    """ Class that creates a simple emitter based on a construct struct.

    For example, if we have a construct format that looks like the following:
        MyStruct = struct(
            "a" / Int8
            "b" / Int8
        )

    We could create emit an object like follows:
        emitter   = ConstructEmitter(MyStruct)
        emitter.a = 0xab
        emitter.b = 0xcd
        my_bytes  = emitter.emit() # "\xab\xcd"
    """

    def __init__(self, struct):
        """
        Parmeters:
            construct_format -- The format for which to create an emitter.
        """
        self.__dict__['format'] = struct
        self.__dict__['fields'] = {}


    def _format_contains_field(self, field_name):
        """ Returns True iff the given format has a field with the provided name.

        Parameters:
            format_object -- The Construct format to work with. This includes e.g. most descriptor types.
            field_name    -- The field name to query.
        """
        return any(f.name == field_name for f in self.format.subcons)


    def __setattr__(self, name, value):
        """ Hook that we used to set our fields. """

        # If the field starts with a '_', don't handle it, as it's an internal field.
        if name.startswith('_'):
            super().__setattr__(name, value)
            return

        if not self._format_contains_field(name):
            raise AttributeError(f"emitter specification contains no field {name}")

        self.fields[name] = value


    def emit(self):
        """ Emits the stream of bytes associated with this object. """

        try:
            return self.format.build(self.fields)
        except KeyError as e:
            raise KeyError(f"missing necessary field: {e}")


    def __getattr__(self, name):
        """ Retrieves an emitter field, if possible. """

        if name in self.fields:
            return self.fields[name]
        else:
            raise AttributeError(f"descriptor emitter has no property {name}")


class ConstructEmitterTest(unittest.TestCase):

    def test_simple_emitter(self):

        test_struct = construct.Struct(
            "a" / construct.Int8ul,
            "b" / construct.Int8ul
        )

        emitter   = ConstructEmitter(test_struct)
        emitter.a = 0xab
        emitter.b = 0xcd

        self.assertEqual(emitter.emit(), b"\xab\xcd")


def emitter_for_format(construct_format):
    """ Creates a factory method for the relevant construct format. """

    def _factory():
        return ConstructEmitter(construct_format)

    return _factory


if __name__ == "__main__":
    unittest.main()



#
# This file is part of usb-protocol.
#
"""
    Descriptors for USB MIDI Class Devices

    [Midi10] refers to "Universal Serial Bus Device Class Definition for MIDI Devices", Release 1.0, November 1, 1999
"""

from enum import IntEnum
import construct

from ..           import USBTransferType, USBSynchronizationType, USBUsageType
from .uac1        import AudioInterfaceClassCode, AudioInterfaceSubclassCodes, AudioClassSpecificDescriptorTypes, AudioClassSpecificACInterfaceDescriptorSubtypes
from ..           import DescriptorTypes
from ..descriptor import DescriptorFormat, DescriptorField, DescriptorNumber

class MidiStreamingInterfaceDescriptorSubtypes(IntEnum):
    # As defined in [Midi10], A.1
    MS_DESCRIPTOR_UNDEFINED = 0x00
    MS_HEADER               = 0x01
    MIDI_IN_JACK            = 0x02
    MIDI_OUT_JACK           = 0x03
    ELEMENT                 = 0x04


class MidiStreamingEndpointDescriptorSubtypes(IntEnum):
    # As defined in [Midi10], A.2
    DESCRIPTOR_UNDEFINED = 0x00
    MS_GENERAL           = 0x01


class MidiStreamingJackTypes(IntEnum):
    # As defined in [Midi10], A.3
    JACK_TYPE_UNDEFINED = 0x00
    EMBEDDED            = 0x01
    EXTERNAL            = 0x02


# As defined in [Midi10], Table 6-1
StandardMidiStreamingInterfaceDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(9, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(DescriptorTypes.INTERFACE),
    "bInterfaceNumber"    / DescriptorField(description="ID of the streaming interface"),
    "bAlternateSetting"   / DescriptorField(description="alternate setting number for the interface", default=0),
    "bNumEndpoints"       / DescriptorField(description="Number of data endpoints used (excluding endpoint 0). Can be: 0 (no data endpoint); 1 (data endpoint); 2 (data + explicit feedback endpoint)", default=0),
    "bInterfaceClass"     / DescriptorNumber(AudioInterfaceClassCode.AUDIO),
    "bInterfaceSubClass"  / DescriptorNumber(AudioInterfaceSubclassCodes.MIDI_STREAMING),
    "bInterfaceProtocol"  / DescriptorNumber(0),
    "iInterface"          / DescriptorField(description="index of a string descriptor describing this interface (0 = unused)", default=0)
)

# As defined in [Midi10], Table 6-2
ClassSpecificMidiStreamingInterfaceHeaderDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(7, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(AudioClassSpecificDescriptorTypes.CS_INTERFACE),
    "bDescriptorSubtype"  / DescriptorNumber(AudioClassSpecificACInterfaceDescriptorSubtypes.HEADER),
    "bcdADC"              / DescriptorField(description="Midi Streaming Class specification release version", default=1.0),
    "wTotalLength"        / DescriptorField(description="Total number of bytes of the class-specific MIDIStreaming interface descriptor. Includes the combined length of this descriptor header and all Jack and Element descriptors."),
)

# As defined in [Midi10], Table 6-3
MidiInJackDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(6, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(AudioClassSpecificDescriptorTypes.CS_INTERFACE),
    "bDescriptorSubtype"  / DescriptorNumber(MidiStreamingInterfaceDescriptorSubtypes.MIDI_IN_JACK),
    "bJackType"           / DescriptorField(description="see MidiStreamingJackTypes"),
    "bJackID"             / DescriptorField(description="Constant uniquely identifying the MIDI IN Jack within the USB-MIDI function"),
    "iJack"               / DescriptorField(description="index of a string descriptor describing this jack (0 = unused)", default=0)
)

# As defined in [Midi10], Table 6-4
MidiOutJackDescriptorHead = DescriptorFormat(
    "bLength"             / DescriptorField(description="Size of this descriptor, in bytes: 6+2*p"),
    "bDescriptorType"     / DescriptorNumber(AudioClassSpecificDescriptorTypes.CS_INTERFACE),
    "bDescriptorSubtype"  / DescriptorNumber(MidiStreamingInterfaceDescriptorSubtypes.MIDI_OUT_JACK),
    "bJackType"           / DescriptorField(description="see MidiStreamingJackTypes"),
    "bJackID"             / DescriptorField(description="Constant uniquely identifying the MIDI IN Jack within the USB-MIDI function"),
    "bNrInputPins"        / DescriptorField(description="Number of Input Pins of this MIDI OUT Jack: p", default=1)
)

MidiOutJackDescriptorElement = DescriptorFormat(
    "baSourceID"          / construct.Int8ul, # ID of the Entity to which the first Input Pin of this MIDI OUT Jack is connected
    "BaSourcePin"         / construct.Int8ul, #Output Pin number of the Entity to which the first Input Pin of this MIDI OUT Jack is connected
)

MidiOutJackDescriptorFoot = DescriptorFormat(
    "iJack"               / DescriptorField(description="index of a string descriptor describing this jack (0 = unused)", default=0)
)

# As defined in [Midi10], Table 6-6
StandardMidiStreamingBulkDataEndpointDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(9, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(DescriptorTypes.ENDPOINT),
    "bEndpointAddress"    / DescriptorField(description="The address of the endpoint, use USBDirection.*.from_endpoint_address()"),
    "bmAttributes"        / DescriptorField(description="D1..0: transfer type (10=bulk), D3..2: synchronization type (00=no sync);", default=USBTransferType.BULK | USBSynchronizationType.NONE | USBUsageType.DATA),
    "wMaxPacketSize"      / DescriptorField(description="Maximum packet size this endpoint is capable of", default=512),
    "bInterval"           / DescriptorField(description="Interval for polling endpoint for data transfers expressed in milliseconds. This field is ignored for bulk endpoints. Must be set to 0", default=0),
    "bRefresh"            / DescriptorField(description="must be set to 0", default=0),
    "bSynchAddress"       / DescriptorField(description="The address of the endpoint used to communicate synchronization information if required by this endpoint. Must be set to 0", default=0)
)

# As defined in [Midi10], Table 6-7
ClassSpecificMidiStreamingBulkDataEndpointDescriptorHead = DescriptorFormat(
    "bLength"             / DescriptorField(description="Size of this descriptor, in bytes: 4+n"),
    "bDescriptorType"     / DescriptorNumber(AudioClassSpecificDescriptorTypes.CS_ENDPOINT),
    "bDescriptorSubtype"  / DescriptorField(description="see MidiStreamingEndpointDescriptorSubtypes", default=MidiStreamingEndpointDescriptorSubtypes.MS_GENERAL),
    "bNumEmbMIDIJack"     / DescriptorField(description="Number of Embedded MIDI Jacks: n", default=1)
)

ClassSpecificMidiStreamingBulkDataEndpointDescriptorElement = DescriptorFormat(
    "baAssocJackID"       / construct.Int8ul # ID of the embedded eack that is associated with this endpoint
)



#
# This file is part of usb_protocol.
#
""" Convenience emitters for USB Audio Class 1 descriptors. """

from contextlib import contextmanager

from .. import emitter_for_format
from ...types.descriptors.uac1 import *

AudioControlInterruptEndpointDescriptorEmitter  = emitter_for_format(AudioControlInterruptEndpointDescriptor)



#
# This file is part of usb-protocol.
#
"""
Structures describing standard USB descriptors. Versions that support parsing incomplete binary data
are available as `DescriptorType`.Partial, e.g. `DeviceDescriptor.Partial`, and are collectively available
in the `usb_protocol.types.descriptors.partial.standard` module (which, like the structs in this module,
can also be imported without `.standard`).
"""

import unittest
from enum import IntEnum

import construct
from   construct  import this, Default

from .. import LanguageIDs
from ..descriptor import \
    DescriptorField, DescriptorNumber, DescriptorFormat, \
    BCDFieldAdapter, DescriptorLength


class StandardDescriptorNumbers(IntEnum):
    """ Numbers of our standard descriptors. """

    DEVICE                                        =  1
    CONFIGURATION                                 =  2
    STRING                                        =  3
    INTERFACE                                     =  4
    ENDPOINT                                      =  5
    DEVICE_QUALIFIER                              =  6
    OTHER_SPEED_DESCRIPTOR                        =  7
    OTHER_SPEED                                   =  7
    INTERFACE_POWER                               =  8
    OTG                                           =  9
    DEBUG                                         = 10
    INTERFACE_ASSOCIATION                         = 11

    # SuperSpeed only
    BOS                                           = 15
    DEVICE_CAPABILITY                             = 16
    SUPERSPEED_USB_ENDPOINT_COMPANION             = 48
    SUPERSPEEDPLUS_ISOCHRONOUS_ENDPOINT_COMPANION = 49


class DeviceCapabilityTypes(IntEnum):
    """ Numbers for the SuperSpeed standard Device Capabilities. """

    WIRELESS_USB                =  1
    USB_2_EXTENSION             =  2
    SUPERSPEED_USB              =  3
    CONTAINER_ID                =  4
    PLATFORM                    =  5
    POWER_DELIVERY_CAPABILITY   =  6
    BATTERY_INFO_CAPABILITY     =  7
    PD_CONSUMER_PORT_CAPABILITY =  8
    PD_PROVIDER_PORT_CAPABILITY =  9
    SUPERSPEED_PLUS             = 10
    PRECISION_TIME_MEASUREMENT  = 11
    WIRELESS_USB_EXTENSION      = 12
    BILLBOARD                   = 13
    AUTHENTICATION              = 14
    BILLBOARD_EXTENSION         = 15
    CONFIGURATION_SUMMARY       = 16



DeviceDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(0x12, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(StandardDescriptorNumbers.DEVICE),
    "bcdUSB"              / DescriptorField("USB Version", default=2.0),
    "bDeviceClass"        / DescriptorField("Class",    default=0),
    "bDeviceSubclass"     / DescriptorField("Subclass", default=0),
    "bDeviceProtocol"     / DescriptorField("Protocol", default=0),
    "bMaxPacketSize0"     / DescriptorField("EP0 Max Pkt Size", default=64),
    "idVendor"            / DescriptorField("Vendor ID"),
    "idProduct"           / DescriptorField("Product ID"),
    "bcdDevice"           / DescriptorField("Device Version", default=0),
    "iManufacturer"       / DescriptorField("Manufacturer Str", default=0),
    "iProduct"            / DescriptorField("Product Str", default=0),
    "iSerialNumber"       / DescriptorField("Serial Number", default=0),
    "bNumConfigurations"  / DescriptorField("Configuration Count"),
)



ConfigurationDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(9, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(StandardDescriptorNumbers.CONFIGURATION),
    "wTotalLength"        / DescriptorField("Length including subordinates"),
    "bNumInterfaces"      / DescriptorField("Interface count"),
    "bConfigurationValue" / DescriptorField("Configuration number", default=1),
    "iConfiguration"      / DescriptorField("Description string", default=0),
    "bmAttributes"        / DescriptorField("Attributes", default=0x80),
    "bMaxPower"           / DescriptorField("Max power consumption", default=250),
)

# Field that automatically reflects a string descriptor's length.
StringDescriptorLength = construct.Rebuild(construct.Int8ul, construct.len_(this.bString) * 2 + 2)

StringDescriptor = DescriptorFormat(
    "bLength"             / StringDescriptorLength,
    "bDescriptorType"     / DescriptorNumber(StandardDescriptorNumbers.STRING),
    "bString"             / construct.GreedyString("utf_16_le")
)


StringLanguageDescriptorLength = \
     construct.Rebuild(construct.Int8ul, construct.len_(this.wLANGID) * 2 + 2)

StringLanguageDescriptor = DescriptorFormat(
    "bLength"             / StringLanguageDescriptorLength,
    "bDescriptorType"     / DescriptorNumber(StandardDescriptorNumbers.STRING),
    "wLANGID"             / construct.GreedyRange(construct.Int16ul)
)


InterfaceDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(9, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(StandardDescriptorNumbers.INTERFACE),
    "bInterfaceNumber"    / DescriptorField("Interface number"),
    "bAlternateSetting"   / DescriptorField("Alternate setting", default=0),
    "bNumEndpoints"       / DescriptorField("Endpoints included"),
    "bInterfaceClass"     / DescriptorField("Class", default=0xff),
    "bInterfaceSubclass"  / DescriptorField("Subclass", default=0xff),
    "bInterfaceProtocol"  / DescriptorField("Protocol", default=0xff),
    "iInterface"          / DescriptorField("String index", default=0),
)


EndpointDescriptor = DescriptorFormat(

    # [USB2.0: 9.6; USB Audio Device Class Definition 1.0: 4.6.1.1, 4.6.2.1]
    # Interfaces of the Audio 1.0 class extend their subordinate endpoint descriptors with
    # 2 additional bytes (extending it from 7 to 9 bytes). Thankfully, this is the only extension that
    # changes the length of a standard descriptor type, but we do have to handle this case in Construct.
    "bLength"             / construct.Default(construct.OneOf(construct.Int8ul, [7, 9]), 7),
    "bDescriptorType"     / DescriptorNumber(StandardDescriptorNumbers.ENDPOINT),
    "bEndpointAddress"    / DescriptorField("Endpoint Address"),
    "bmAttributes"        / DescriptorField("Attributes", default=2),
    "wMaxPacketSize"      / DescriptorField("Maximum Packet Size", default=64),
    "bInterval"           / DescriptorField("Polling interval", default=255),

    # 2 bytes that are only present on endpoint descriptors for Audio 1.0 class interfaces.
    ("bRefresh"           / construct.Optional(construct.Int8ul)) * "Refresh Rate",
    ("bSynchAddress"      / construct.Optional(construct.Int8ul)) * "Synch Endpoint Address",
)


DeviceQualifierDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(9, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(StandardDescriptorNumbers.DEVICE_QUALIFIER),
    "bcdUSB"              / DescriptorField("USB Version"),
    "bDeviceClass"        / DescriptorField("Class"),
    "bDeviceSubclass"     / DescriptorField("Subclass"),
    "bDeviceProtocol"     / DescriptorField("Protocol"),
    "bMaxPacketSize0"     / DescriptorField("EP0 Max Pkt Size"),
    "bNumConfigurations"  / DescriptorField("Configuration Count"),
    "_bReserved"          / construct.Optional(construct.Const(b"\0"))
)


#
# SuperSpeed descriptors
#
BinaryObjectStoreDescriptor = DescriptorFormat(
    "bLength"               / construct.Const(0x5, construct.Int8ul),
    "bDescriptorType"       / DescriptorNumber(StandardDescriptorNumbers.BOS),
    "wTotalLength"          / DescriptorField("Total Length", default=5),
    "bNumDeviceCaps"        / DescriptorField("Device Capability Descriptors", default=0),
)

USB2ExtensionDescriptor = DescriptorFormat(
    "bLength"               / construct.Const(0x7, construct.Int8ul),
    "bDescriptorType"       / DescriptorNumber(StandardDescriptorNumbers.DEVICE_CAPABILITY),
    "bDevCapabilityType"    / construct.Const(DeviceCapabilityTypes.USB_2_EXTENSION, construct.Int8ul),
    "bmAttributes"          / DescriptorField("Attributes", default=0b10, length=4)
)

SuperSpeedUSBDeviceCapabilityDescriptor = DescriptorFormat(
    "bLength"               / construct.Const(0xA, construct.Int8ul),
    "bDescriptorType"       / DescriptorNumber(StandardDescriptorNumbers.DEVICE_CAPABILITY),
    "bDevCapabilityType"    / construct.Const(DeviceCapabilityTypes.SUPERSPEED_USB, construct.Int8ul),
    "bmAttributes"          / DescriptorField("Attributes", default=0),
    "wSpeedsSupported"      / DescriptorField("USB3 Speeds Supported", default=0b1000),
    "bFunctionalitySupport" / DescriptorField("Lowest Speed with Full Support", default=3),
    "bU1DevExitLat"         / DescriptorField("U1 Exit Latency", default=0),
    "wU2DevExitLat"         / DescriptorField("U2 Exit Latency", default=0)
)


SuperSpeedEndpointCompanionDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(0x6, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(StandardDescriptorNumbers.SUPERSPEED_USB_ENDPOINT_COMPANION),
    "bMaxBurst"           / DescriptorField("Maximum Burst Length", default=0),
    "bmAttributes"        / DescriptorField("Extended Attributes", default=0),
    "wBytesPerInterval"   / DescriptorField("Bytes Per Service Interval", default=0),
)



class DescriptorParserCases(unittest.TestCase):

    STRING_DESCRIPTOR = bytes([
        40, # Length
        3,  # Type
        ord('G'), 0x00,
        ord('r'), 0x00,
        ord('e'), 0x00,
        ord('a'), 0x00,
        ord('t'), 0x00,
        ord(' '), 0x00,
        ord('S'), 0x00,
        ord('c'), 0x00,
        ord('o'), 0x00,
        ord('t'), 0x00,
        ord('t'), 0x00,
        ord(' '), 0x00,
        ord('G'), 0x00,
        ord('a'), 0x00,
        ord('d'), 0x00,
        ord('g'), 0x00,
        ord('e'), 0x00,
        ord('t'), 0x00,
        ord('s'), 0x00,
    ])


    def test_string_descriptor_parse(self):

        # Parse the relevant string...
        parsed = StringDescriptor.parse(self.STRING_DESCRIPTOR)

        # ... and check the desriptor's fields.
        self.assertEqual(parsed.bLength,                    40)
        self.assertEqual(parsed.bDescriptorType,             3)
        self.assertEqual(parsed.bString, "Great Scott Gadgets")


    def test_string_descriptor_build(self):
        data = StringDescriptor.build({
            'bString': "Great Scott Gadgets"
        })

        self.assertEqual(data, self.STRING_DESCRIPTOR)


    def test_string_language_descriptor_build(self):
        data = StringLanguageDescriptor.build({
            'wLANGID': (LanguageIDs.ENGLISH_US,)
        })

        self.assertEqual(data, b"\x04\x03\x09\x04")


    def test_device_descriptor(self):

        device_descriptor = [
            0x12,         # Length
            0x01,         # Type
            0x00, 0x02,   # USB version
            0xFF,         # class
            0xFF,         # subclass
            0xFF,         # protocol
            64,           # ep0 max packet size
            0xd0, 0x16,   # VID
            0x3b, 0x0f,   # PID
            0x00, 0x00,   # device rev
            0x01,         # manufacturer string
            0x02,         # product string
            0x03,         # serial number
            0x01          # number of configurations
        ]

        # Parse the relevant string...
        parsed = DeviceDescriptor.parse(device_descriptor)

        # ... and check the desriptor's fields.
        self.assertEqual(parsed.bLength,             18)
        self.assertEqual(parsed.bDescriptorType,      1)
        self.assertEqual(parsed.bcdUSB,             2.0)
        self.assertEqual(parsed.bDeviceClass,      0xFF)
        self.assertEqual(parsed.bDeviceSubclass,   0xFF)
        self.assertEqual(parsed.bDeviceProtocol,   0xFF)
        self.assertEqual(parsed.bMaxPacketSize0,     64)
        self.assertEqual(parsed.idVendor,        0x16d0)
        self.assertEqual(parsed.idProduct,       0x0f3b)
        self.assertEqual(parsed.bcdDevice,            0)
        self.assertEqual(parsed.iManufacturer,        1)
        self.assertEqual(parsed.iProduct,             2)
        self.assertEqual(parsed.iSerialNumber,        3)
        self.assertEqual(parsed.bNumConfigurations,   1)


    def test_bcd_constructor(self):

        emitter = BCDFieldAdapter(construct.Int16ul)
        result = emitter.build(1.4)

        self.assertEqual(result, b"\x40\x01")


if __name__ == "__main__":
    unittest.main()



#
# This file is part of usb-protocol.
#
"""
    Descriptors for USB MIDI Class Devices

    [Midi20] refers to "Universal Serial Bus Device Class Definition for MIDI Devices", Release 2.0, May 5, 2020
"""

from enum import IntEnum

import construct

from .uac2        import AudioInterfaceClassCode, AudioInterfaceSubclassCodes, AudioClassSpecificDescriptorTypes, AudioClassSpecificACInterfaceDescriptorSubtypes
from ..           import DescriptorTypes
from ..descriptor import DescriptorFormat, DescriptorField, DescriptorNumber



class MidiStreamingInterfaceDescriptorTypes(IntEnum):
    # As defined in [Midi20], A.1
    CS_UNDEFINED     = 0x20
    CS_DEVICE        = 0x21
    CS_CONFIGURATION = 0x22
    CS_STRING        = 0x23
    CS_INTERFACE     = 0x24
    CS_ENDPOINT      = 0x25
    CS_GR_TRM_BLOCK  = 0x26


class MidiStreamingInterfaceDescriptorSubtypes(IntEnum):
    # As defined in [Midi20], A.1
    MS_DESCRIPTOR_UNDEFINED = 0x00
    MS_HEADER               = 0x01
    MIDI_IN_JACK            = 0x02
    MIDI_OUT_JACK           = 0x03
    ELEMENT                 = 0x04


class MidiStreamingEndpointDescriptorSubtypes(IntEnum):
    # As defined in [Midi20], A.2
    DESCRIPTOR_UNDEFINED = 0x00
    MS_GENERAL           = 0x01
    MS_GENERAL_2_0       = 0x02


class MidiStreamingInterfaceHeaderClassRevision(IntEnum):
    # As defined in [Midi20], A.4
    MS_MIDI_1_0 = 0x0100
    MS_MIDI_2_0 = 0x0200


class MidiStreamingJackTypes(IntEnum):
    # As defined in [Midi20], A.5
    JACK_TYPE_UNDEFINED = 0x00
    EMBEDDED            = 0x01
    EXTERNAL            = 0x02


# As defined in [Midi20], Table B-5
StandardMidiStreamingInterfaceDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(9, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(DescriptorTypes.INTERFACE),
    "bInterfaceNumber"    / DescriptorField(description="ID of the streaming interface"),
    "bAlternateSetting"   / DescriptorField(description="alternate setting number for the interface", default=0),
    "bNumEndpoints"       / DescriptorField(description="Number of data endpoints used (excluding endpoint 0). Can be: 0 (no data endpoint); 1 (data endpoint); 2 (data + explicit feedback endpoint)", default=0),
    "bInterfaceClass"     / DescriptorNumber(AudioInterfaceClassCode.AUDIO),
    "bInterfaceSubClass"  / DescriptorNumber(AudioInterfaceSubclassCodes.MIDI_STREAMING),
    "bInterfaceProtocol"  / DescriptorNumber(0),
    "iInterface"          / DescriptorField(description="index of a string descriptor describing this interface (0 = unused)", default=0)
)

# As defined in [Midi20], Table B-6
ClassSpecificMidiStreamingInterfaceHeaderDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(7, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(AudioClassSpecificDescriptorTypes.CS_INTERFACE),
    "bDescriptorSubtype"  / DescriptorNumber(AudioClassSpecificACInterfaceDescriptorSubtypes.HEADER),
    "bcdADC"              / DescriptorField(description="Midi Streaming Class specification release version", default=1.0),
    "wTotalLength"        / DescriptorField(description="Total number of bytes of the class-specific MIDIStreaming interface descriptor. Includes the combined length of this descriptor header and all Jack and Element descriptors."),
)

# As defined in [Midi20], Table 5-3
StandardMidiStreamingDataEndpointDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(7, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(AudioClassSpecificDescriptorTypes.CS_ENDPOINT),
    "bEndpointAddress"    / DescriptorField(description="endpoint address, use USBDirection.*.from_endpoint_address()"),
    "bmAttributes"        / DescriptorField(description="endpoint type, see USBTransferType (only NONE, BULK or INTERRUPT allowed)", default=USBTransferType.BULK),
    "wMaxPacketSize"      / DescriptorField(description="Maximum packet size this endpoint is capable of sending or receiving"),
    "bInterval"           / DescriptorField(description="Interval for polling endpoint for Interrupt data transfers. For bulk endpoints this field is ignored and must be reset to 0", default=0)
)






#
# This file is part of usb-protocol.
#
"""
    Descriptors for USB Audio Class Devices (UAC), Release 1

    [Audio10] refers to "Universal Serial Bus Device Class Definition for Audio Devices", Release 1.0, March 18, 1998
    [Frmts10] refers to "Universal Serial Bus Device Class Definition for Audio Data Formats", Release 1.0, March 18, 1998
    [TermT10] refers to "Universal Serial Bus Device Class Definition for Terminal Types", Release 1.0, March 18, 1998
"""

from enum import IntEnum
import construct

from ..descriptor import DescriptorFormat, DescriptorField, DescriptorNumber


class AudioInterfaceClassCode(IntEnum):
    # As defined in [Audio10], Table A-1
    AUDIO = 0x01


class AudioInterfaceSubclassCodes(IntEnum):
    # As defined in [Audio10], Table A-2
    INTERFACE_SUBCLASS_UNDEFINED = 0x00
    AUDIO_CONTROL                = 0x01
    AUDIO_STREAMING              = 0x02
    MIDI_STREAMING               = 0x03


class AudioInterfaceProtocolCodes(IntEnum):
    # As defined in [Audio10], Table A-3
    PR_PROTOCOL_UNDEFINED = 0x00


class AudioClassSpecificDescriptorTypes(IntEnum):
    # As defined in [Audio10], Table A-4
    CS_UNDEFINED     = 0x20
    CS_DEVICE        = 0x21
    CS_CONFIGURATION = 0x22
    CS_STRING        = 0x23
    CS_INTERFACE     = 0x24
    CS_ENDPOINT      = 0x25


class AudioClassSpecificACInterfaceDescriptorSubtypes(IntEnum):
    # As defined in [Audio10], Table A-5
    AC_DESCRIPTOR_UNDEFINED  = 0x00
    HEADER                   = 0x01
    INPUT_TERMINAL           = 0x02
    OUTPUT_TERMINAL          = 0x03
    MIXER_UNIT               = 0x04
    SELECTOR_UNIT            = 0x05
    FEATURE_UNIT             = 0x06
    PROCESSING_UNIT          = 0x07
    EXTENSION_UNIT           = 0x08


class AudioClassSpecificASInterfaceDescriptorSubtypes(IntEnum):
    # As defined in [Audio10], Table A-6
    AS_DESCRIPTOR_UNDEFINED = 0x00
    AS_GENERAL              = 0x01
    FORMAT_TYPE             = 0x02
    FORMAT_SPECIFIC         = 0x03


class ProcessingUnitProcessTypes(IntEnum):
    # As defined in [Audio10], Table A-7
    PROCESS_UNDEFINED           = 0x00
    UP_DOWNMIX_PROCESS          = 0x01
    DOLBY_PROLOGIC_PROCESS      = 0x02
    _3D_STEREO_EXTENDER_PROCESS = 0x03
    REVERBERATION_PROCESS       = 0x04
    CHORUS_PROCESS              = 0x05
    DYN_RANGE_COMP_PROCESS      = 0x06


class AudioClassSpecificEndpointDescriptorSubtypes(IntEnum):
    # As defined in [Audio10], Table A-8
    DESCRIPTOR_UNDEFINED = 0x00
    EP_GENERAL           = 0x01


class AudioClassSpecificRequestCodes(IntEnum):
    # As defined in [Audio10], Table A-9
    REQUEST_CODE_UNDEFINED = 0x00
    SET_CUR                = 0x01
    GET_CUR                = 0x81
    SET_MIN                = 0x02
    GET_MIN                = 0x82
    SET_MAX                = 0x03
    GET_MAX                = 0x83
    SET_RES                = 0x04
    GET_RES                = 0x84
    SET_MEM                = 0x05
    GET_MEM                = 0x85
    GET_STAT               = 0xFF


class TerminalControlSelectors(IntEnum):
    # As defined in [Audio10], Table A-10
    TE_CONTROL_UNDEFINED = 0x00
    COPY_PROTECT_CONTROL = 0x01


class FeatureUnitControlSelectors(IntEnum):
    # As defined in [Audio10], Table A-11
    FU_CONTROL_UNDEFINED      = 0x00
    MUTE_CONTROL              = 0x01
    VOLUME_CONTROL            = 0x02
    BASS_CONTROL              = 0x03
    MID_CONTROL               = 0x04
    TREBLE_CONTROL            = 0x05
    GRAPHIC_EQUALIZER_CONTROL = 0x06
    AUTOMATIC_GAIN_CONTROL    = 0x07
    DELAY_CONTROL             = 0x08
    BASS_BOOST_CONTROL        = 0x09
    LOUDNESS_CONTROL          = 0x0A


class UpDownMixProcessingUnitControlSelectors(IntEnum):
    # As defined in [Audio10], Table A-12
    UD_CONTROL_UNDEFINED   = 0x00
    UD_ENABLE_CONTROL      = 0x01
    UD_MODE_SELECT_CONTROL = 0x02


class DolbyProLogicProcessingUnitControlSelectors(IntEnum):
    # As defined in [Audio10], Table A-13
    DP_CONTROL_UNDEFINED   = 0x00
    DP_ENABLE_CONTROL      = 0x01
    DP_MODE_SELECT_CONTROL = 0x02


class _3DStereoExtenderProcessingUnitControlSelectors(IntEnum):
    # As defined in [Audio10], Table A-14
    _3D_CONTROL_UNDEFINED = 0x00
    _3D_ENABLE_CONTROL    = 0x01
    SPACIOUSNESS_CONTROL  = 0x02


class ReverberationProcessingUnitControlSelectors(IntEnum):
    # As defined in [Audio10], Table A-15
    RV_CONTROL_UNDEFINED    = 0x00
    RV_ENABLE_CONTROL       = 0x01
    REVERB_LEVEL_CONTROL    = 0x02
    REVERB_TIME_CONTROL     = 0x03
    REVERB_FEEDBACK_CONTROL = 0x04


class ChorusProcessingUnitControlSelectors(IntEnum):
    # As defined in [Audio10], Table A-16
    CH_CONTROL_UNDEFINED = 0x00
    CH_ENABLE_CONTROL    = 0x01
    CHORUS_LEVEL_CONTROL = 0x02
    CHORUS_RATE_CONTROL  = 0x03
    CHORUS_DEPTH_CONTROL = 0x04


class DynamicRangeCompressorProcessingUnitControlSelectors(IntEnum):
    # As defined in [Audio10], Table A-17
    DR_CONTROL_UNDEFINED     = 0x00
    DR_ENABLE_CONTROL        = 0x01
    COMPRESSION_RATE_CONTROL = 0x02
    MAXAMPL_CONTROL          = 0x03
    THRESHOLD_CONTROL        = 0x04
    ATTACK_TIME              = 0x05
    RELEASE_TIME             = 0x06


class ExtensionUnitControlSelectors(IntEnum):
    # As defined in [Audio10], Table A-18
    XU_CONTROL_UNDEFINED     = 0x00
    XU_ENABLE_CONTROL        = 0x01


class EndpointsControlSelectors(IntEnum):
    # As defined in [Audio10], Table A-19
    EP_CONTROL_UNDEFINED  = 0x00
    SAMPLING_FREQ_CONTROL = 0x01
    PITCH_CONTROL         = 0x02


class USBTerminalTypes(IntEnum):
    # As defined in [TermT10], Table 2-1
    USB_UNDEFINED       = 0x0100
    USB_STREAMING       = 0x0101
    USB_VENDOR_SPECIFIC = 0x01FF


class InputTerminalTypes(IntEnum):
    # As defined in [TermT10], Table 2-2
    INPUT_UNDEFINED             = 0x0200
    MICROPHONE                  = 0x0201
    DESKTOP_MICROPHONE          = 0x0202
    PERSONAL_MICROPHONE         = 0x0203
    OMNI_DIRECTIONAL_MICROPHONE = 0x0204
    MICROPHONE_ARRAY            = 0x0205
    PROCESSING_MICROPHONE_ARRAY = 0x0206


class OutputTerminalTypes(IntEnum):
    # As defined in [TermT10], Table 2-3
    OUTPUT_UNDEFINED              = 0x0300
    SPEAKER                       = 0x0301
    HEADPHONES                    = 0x0302
    DESKTOP_SPEAKER               = 0x0304
    ROOM_SPEAKER                  = 0x0305
    COMMUNICATION_SPEAKER         = 0x0306
    LOW_FREQUENCY_EFFECTS_SPEAKER = 0x0307


class BidirectionalTerminalTypes(IntEnum):
    # As defined in [TermT10], Table 2-4
    BIDIRECTIONAL_UNDEFINED       = 0x0400
    HANDSET                       = 0x0401
    HEADSET                       = 0x0402
    ECHO_SUPPRESSING_SPEAKERPHONE = 0x0404
    ECHO_CANCELING_SPEAKERPHONE   = 0x0405


class TelephonyTerminalTypes(IntEnum):
    # As defined in [TermT10], Table 2-5
    TELEPHONY_UNDEFINED = 0x0500
    PHONE_LINE          = 0x0501
    TELEPHONE           = 0x0502
    DOWN_LINE_PHONE     = 0x0503


class ExternalTerminalTypes(IntEnum):
    # As defined in [TermT10], Table 2-6
    EXTERNAL_UNDEFINED             = 0x0600
    ANALOG_CONNECTOR               = 0x0601
    DIGITAL_AUDIO_INTERFACE        = 0x0602
    LINE_CONNECTOR                 = 0x0603
    SPDIF_INTERFACE                = 0x0605
    IEEE_1394_DA_STREAM            = 0x0606
    IEEE_1394_DV_STREAM_SOUNDTRACK = 0x0607


class EmbeddedFunctionTerminalTypes(IntEnum):
    # As defined in [TermT10], Table 2-7
    EMBEDDED_UNDEFINED       = 0x0700
    EQUALIZATION_NOISE       = 0x0702
    CD_PLAYER                = 0x0703
    DAT                      = 0x0704
    DCC                      = 0x0705
    ANALOG_TAPE              = 0x0707
    PHONOGRAPH               = 0x0708
    VCR_AUDIO                = 0x0709
    VIDEO_DISC_AUDIO         = 0x070A
    DVD_AUDIO                = 0x070B
    TV_TUNER_AUDIO           = 0x070C
    SATELLITE_RECEIVER_AUDIO = 0x070D
    CABLE_TUNER_AUDIO        = 0x070E
    DSS_AUDIO                = 0x070F
    RADIO_RECEIVER           = 0x0710
    RADIO_TRANSMITTER        = 0x0711
    MULTI_TRACK_RECORDER     = 0x0712
    SYNTHESIZER              = 0x0713


# As defined in [Audio10], Table 4-17
AudioControlInterruptEndpointDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(9, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(AudioClassSpecificDescriptorTypes.CS_ENDPOINT),
    "bEndpointAddress"    / DescriptorField(description="The address of the endpoint, use USBDirection.*.from_endpoint_address()"),
    "bmAttributes"        / DescriptorField(description="D1..0: Transfer type (0b11 = Interrupt)", default=0b11),
    "wMaxPacketSize"      / DescriptorField(description="Maximum packet size this endpoint is capable of. Used here to pass 6-byte interrupt information.", default=6),
    "bInterval"           / DescriptorField(description="Interval for polling the Interrupt endpoint"),
    "bRefresh"            / DescriptorField(description="Reset to 0"),
    "bSynchAddress"       / DescriptorField(description="Reset to 0")
)




#
# This file is part of usb_protocol.
#
""" Convenience emitters for USB MIDI Class 2 descriptors. """

from .. import emitter_for_format
from ...types.descriptors.midi2 import *

StandardMidiStreamingInterfaceDescriptorEmitter            = emitter_for_format(StandardMidiStreamingInterfaceDescriptor)
ClassSpecificMidiStreamingInterfaceHeaderDescriptorEmitter = emitter_for_format(ClassSpecificMidiStreamingInterfaceHeaderDescriptor)
StandardMidiStreamingDataEndpointDescriptorEmitter         = emitter_for_format(StandardMidiStreamingDataEndpointDescriptor)


#
# This file is part of usb_protocol.
#
""" Convenience emitters for CDC descriptors. """

from .. import emitter_for_format
from ...types.descriptors.cdc import \
    HeaderDescriptor, UnionFunctionalDescriptor, ACMFunctionalDescriptor, \
    CallManagementFunctionalDescriptor

# Create our emitters.
HeaderDescriptorEmitter                   = emitter_for_format(HeaderDescriptor)
UnionFunctionalDescriptorEmitter          = emitter_for_format(UnionFunctionalDescriptor)
ACMFunctionalDescriptorEmitter            = emitter_for_format(ACMFunctionalDescriptor)
CallManagementFunctionalDescriptorEmitter = emitter_for_format(CallManagementFunctionalDescriptor)



#
# This file is part of usb_protocol.
#
""" Convenience emitters for USB Audio Class 2 descriptors. """

from contextlib import contextmanager

from .. import emitter_for_format
from ...types.descriptors.uac2 import *
from ...emitters.descriptor    import ComplexDescriptorEmitter

# Create our emitters.
InterfaceAssociationDescriptorEmitter          = emitter_for_format(InterfaceAssociationDescriptor)
StandardAudioControlInterfaceDescriptorEmitter = emitter_for_format(StandardAudioControlInterfaceDescriptor)

class ClassSpecificAudioControlInterfaceDescriptorEmitter(ComplexDescriptorEmitter):
    DESCRIPTOR_FORMAT = ClassSpecificAudioControlInterfaceDescriptor

    def _pre_emit(self):
        # Figure out the total length of our descriptor, including subordinates.
        subordinate_length = sum(len(sub) for sub in self._subordinates)
        self.wTotalLength = subordinate_length + self.DESCRIPTOR_FORMAT.sizeof()

ClockSourceDescriptorEmitter                                             = emitter_for_format(ClockSourceDescriptor)
InputTerminalDescriptorEmitter                                           = emitter_for_format(InputTerminalDescriptor)
OutputTerminalDescriptorEmitter                                          = emitter_for_format(OutputTerminalDescriptor)
AudioStreamingInterfaceDescriptorEmitter                                 = emitter_for_format(AudioStreamingInterfaceDescriptor)
ClassSpecificAudioStreamingInterfaceDescriptorEmitter                    = emitter_for_format(ClassSpecificAudioStreamingInterfaceDescriptor)
TypeIFormatTypeDescriptorEmitter                                         = emitter_for_format(TypeIFormatTypeDescriptor)
ExtendedTypeIFormatTypeDescriptorEmitter                                 = emitter_for_format(ExtendedTypeIFormatTypeDescriptor)
TypeIIFormatTypeDescriptorEmitter                                        = emitter_for_format(TypeIIFormatTypeDescriptor)
ExtendedTypeIIFormatTypeDescriptorEmitter                                = emitter_for_format(ExtendedTypeIIFormatTypeDescriptor)
TypeIIIFormatTypeDescriptorEmitter                                       = emitter_for_format(TypeIIIFormatTypeDescriptor)
ExtendedTypeIIIFormatTypeDescriptorEmitter                               = emitter_for_format(ExtendedTypeIIIFormatTypeDescriptor)
ClassSpecificAudioStreamingIsochronousAudioDataEndpointDescriptorEmitter = emitter_for_format(ClassSpecificAudioStreamingIsochronousAudioDataEndpointDescriptor)
AudioControlInterruptEndpointDescriptorEmitter                           = emitter_for_format(AudioControlInterruptEndpointDescriptor)
AudioStreamingIsochronousEndpointDescriptorEmitter                       = emitter_for_format(AudioStreamingIsochronousEndpointDescriptor)
AudioStreamingIsochronousFeedbackEndpointDescriptorEmitter               = emitter_for_format(AudioStreamingIsochronousFeedbackEndpointDescriptor)


#
# This file is part of usb_protocol.
#
""" Convenience emitters for USB MIDI Class 1 descriptors. """

from ..                         import emitter_for_format
from ...emitters.descriptor     import ComplexDescriptorEmitter
from ...types.descriptors.midi1 import *

class ClassSpecificMidiStreamingInterfaceDescriptorEmitter(ComplexDescriptorEmitter):
    DESCRIPTOR_FORMAT = ClassSpecificMidiStreamingInterfaceHeaderDescriptor

    def _pre_emit(self):
        # Figure out the total length of our descriptor, including subordinates.
        subordinate_length = sum(len(sub) for sub in self._subordinates)
        self.wTotalLength = subordinate_length + self.DESCRIPTOR_FORMAT.sizeof()

class MidiOutJackDescriptorEmitter(ComplexDescriptorEmitter):
    DESCRIPTOR_FORMAT = MidiOutJackDescriptorHead

    def add_subordinate_descriptor(self, subordinate):
        subordinate = subordinate.emit()
        self._subordinates.append(subordinate)

    def add_source(self, sourceId, sourcePin=1):
        sourceDescriptor = MidiOutJackDescriptorElementEmitter()
        sourceDescriptor.baSourceID = sourceId
        sourceDescriptor.BaSourcePin = sourcePin
        self.add_subordinate_descriptor(sourceDescriptor)

    def _pre_emit(self):
        self.add_subordinate_descriptor(MidiOutJackDescriptorFootEmitter())
        # Figure out the total length of our descriptor, including subordinates.
        subordinate_length = sum(len(sub) for sub in self._subordinates)
        self.bLength = subordinate_length + self.DESCRIPTOR_FORMAT.sizeof()

class ClassSpecificMidiStreamingBulkDataEndpointDescriptorEmitter(ComplexDescriptorEmitter):
    DESCRIPTOR_FORMAT = ClassSpecificMidiStreamingBulkDataEndpointDescriptorHead

    def add_subordinate_descriptor(self, subordinate):
        subordinate = subordinate.emit()
        self._subordinates.append(subordinate)

    def add_associated_jack(self, jackID):
        jackDescriptor = ClassSpecificMidiStreamingBulkDataEndpointDescriptorElementEmitter()
        jackDescriptor.baAssocJackID = jackID
        self.add_subordinate_descriptor(jackDescriptor)

    def _pre_emit(self):
        # Figure out the total length of our descriptor, including subordinates.
        subordinate_length = sum(len(sub) for sub in self._subordinates)
        self.bLength = subordinate_length + self.DESCRIPTOR_FORMAT.sizeof()

StandardMidiStreamingInterfaceDescriptorEmitter                    = emitter_for_format(StandardMidiStreamingInterfaceDescriptor)
ClassSpecificMidiStreamingInterfaceHeaderDescriptorEmitter         = emitter_for_format(ClassSpecificMidiStreamingInterfaceHeaderDescriptor)
MidiInJackDescriptorEmitter                                        = emitter_for_format(MidiInJackDescriptor)
MidiOutJackDescriptorHeadEmitter                                   = emitter_for_format(MidiOutJackDescriptorHead)
MidiOutJackDescriptorElementEmitter                                = emitter_for_format(MidiOutJackDescriptorElement)
MidiOutJackDescriptorFootEmitter                                   = emitter_for_format(MidiOutJackDescriptorFoot)
StandardMidiStreamingBulkDataEndpointDescriptorEmitter             = emitter_for_format(StandardMidiStreamingBulkDataEndpointDescriptor)
ClassSpecificMidiStreamingBulkDataEndpointDescriptorHeadEmitter    = emitter_for_format(ClassSpecificMidiStreamingBulkDataEndpointDescriptorHead)
ClassSpecificMidiStreamingBulkDataEndpointDescriptorElementEmitter = emitter_for_format(ClassSpecificMidiStreamingBulkDataEndpointDescriptorElement)


#!/usr/bin/env python3
#
# This file is part of usb-protocol.
#
""" Example that builds a device-worth of descriptors using a collection object. """

from usb_protocol.emitters.descriptors import DeviceDescriptorCollection

collection = DeviceDescriptorCollection()

# Create our device descriptor, and populate it with some fields.
# Many fields have sane/common defaults; and thus can be omitted.
with collection.DeviceDescriptor() as d:
    d.idVendor           = 0xc001
    d.idProduct          = 0xc0de
    d.bNumConfigurations = 1

    d.iManufacturer      = "usb-tools"
    d.iProduct           = "Illegitimate USB Device"
    d.iSerialNumber      = "123456"


# Create our configuration descriptor, and its subordinates.
with collection.ConfigurationDescriptor() as c:
    # Here, this is our first configuration, and is automatically assigned number '1'.

    # We'll create a simple interface with a couple of endpoints.
    with c.InterfaceDescriptor() as i:
        i.bInterfaceNumber = 0

        # Our endpoints default to bulk; with mostly-sane defaults.
        with i.EndpointDescriptor() as e:
            e.bEndpointAddress = 0x01
            e.wMaxPacketSize   = 512

        with i.EndpointDescriptor() as e:
            e.bEndpointAddress = 0x81
            e.wMaxPacketSize   = 512


print("This device's descriptors would look like:")

# Iterate over all of our descriptors.
for value, index, raw in collection:
    print(f"    type {value} (index {index}) = {raw}")



""" Convenience aliases for versions of descriptor structs that support parsing incomplete binary data. """

from .. import standard

DeviceDescriptor          = standard.DeviceDescriptor.Partial
ConfigurationDescriptor   = standard.ConfigurationDescriptor.Partial
StringDescriptor          = standard.StringDescriptor.Partial
StringLanguageDescriptor  = standard.StringLanguageDescriptor.Partial
InterfaceDescriptor       = standard.InterfaceDescriptor.Partial
EndpointDescriptor        = standard.EndpointDescriptor.Partial
DeviceQualifierDescriptor = standard.DeviceQualifierDescriptor.Partial



#
# This file is part of usb-protocol.
#


from collections import defaultdict

from . import ConstructEmitter

class ComplexDescriptorEmitter(ConstructEmitter):
    """ Base class for emitting complex descriptors, which contain nested subordinates. """

    # Base classes should override this.
    DESCRIPTOR_FORMAT = None

    def __init__(self, collection=None):
        """
        Parameters:
            collection -- If this descriptor belongs to a collection, it should be
                          provided here. Using a collection object allows e.g. automatic
                          assignment of string descriptor indices.
        """

        self._collection = collection

        # Always create a basic ConstructEmitter from the given format.
        super().__init__(self.DESCRIPTOR_FORMAT)

        # Store a list of subordinate descriptors, and a count of
        # subordinate descriptor types.
        self._subordinates = []
        self._type_counts = defaultdict(int)


    def add_subordinate_descriptor(self, subordinate):
        """ Adds a subordinate descriptor to the relevant descriptor.

        Parameter:
            subordinate -- The subordinate descriptor to add; can be an emitter,
                           or a bytes-like object.
        """

        if hasattr(subordinate, 'emit'):
            subordinate = subordinate.emit()
        else:
            subordinate = bytes(subordinate)

        # The second byte of a given descriptor is always its type number.
        # Count this descriptor type...
        subordinate_type = subordinate[1]

        self._type_counts[subordinate_type] += 1

        # ... and add the relevant bytes to our list of subordinates.
        self._subordinates.append(subordinate)


    def _pre_emit(self):
        """ Performs any manipulations needed on this object before emission. """
        pass


    def emit(self, include_subordinates=True):
        """ Emit our descriptor.

        Parameters:
            include_subordinates -- If true or not provided, any subordinate descriptors will be included.
        """

        # Run any pre-emit hook code before we perform our emission...
        self._pre_emit()

        # Start with our core descriptor...
        result = bytearray()
        result.extend(super().emit())

        # ... and if descired, add our subordinates...
        for sub in self._subordinates:
            result.extend(sub)

        return bytes(result)





#
# This file is part of usb-protocol.
#
""" SuperSpeed USB types -- defines enumerations that describe standard USB3 types. """

from enum import IntEnum

class LinkCommand(IntEnum):
    """ Constant values (including both class and type) for link commands. """

    LGOOD = 0   # Header Packet ACK
    LCRD  = 1   # Header Credit
    LRTY  = 2   # Header Packet Retry Sequence
    LBAD  = 3   # Header Packet NAK
    LGO_U = 4   # Request Switching to Power State Ux
    LAU   = 5   # Power State Acceptance
    LXU   = 6   # Power State Rejection
    LPMA  = 7   # Power State Acknowledgement
    LUP   = 8   # Downstream-facing Keep-alive
    LDN   = 11  # Upstream-facing Keep-alive

    def get_class(self):
        return int(self) >> 2

    def get_type(self):
        return int(self) & 0b11


class HeaderPacketType(IntEnum):
    """ Constants representing the Header Packet archetypes. """
    TRANSACTION           = 0b00100
    DATA                  = 0b01000
    ISOCHRONOUS_TIMESTAMP = 0b01100
    LINK_MANAGEMENT       = 0b00000


class TransactionPacketSubtype(IntEnum):
    """ Constants representing the subtypes of Transition Header Packet. """

    ACK           = 1
    NRDY          = 2
    ERDY          = 3
    STATUS        = 4
    STALL         = 5
    NOTIFICATION  = 6
    PING          = 7
    PING_RESPONSE = 8


class LinkManagementPacketSubtype(IntEnum):
    """ Constants represneting the various types of Link Management Packet. """

    SET_LINK_FUNCTION           = 1
    U2_INACTIVITY_TIMEOUT       = 2
    VENDOR_DEVICE_TEST          = 3
    PORT_CAPABILITY             = 4
    PORT_CONFIGURATION          = 5
    PORT_CONFIGURATION_RESPONSE = 6
    PRECISION_TIME_MANAGEMENT   = 7



#
# This file is part of usb_protocol.
#
""" Convenience emitters for simple, standard descriptors. """

import unittest

from contextlib import contextmanager

from ..           import emitter_for_format
from ..descriptor import ComplexDescriptorEmitter

from ...types     import LanguageIDs
from ...types.descriptors.standard import *


# Create our basic emitters...
DeviceDescriptorEmitter         = emitter_for_format(DeviceDescriptor)
StringDescriptorEmitter         = emitter_for_format(StringDescriptor)
StringLanguageDescriptorEmitter = emitter_for_format(StringLanguageDescriptor)
DeviceQualifierDescriptor       = emitter_for_format(DeviceQualifierDescriptor)

# ... our basic superspeed emitters ...
USB2ExtensionDescriptorEmitter                  = emitter_for_format(USB2ExtensionDescriptor)
SuperSpeedUSBDeviceCapabilityDescriptorEmitter  = emitter_for_format(SuperSpeedUSBDeviceCapabilityDescriptor)
SuperSpeedEndpointCompanionDescriptorEmitter    = emitter_for_format(SuperSpeedEndpointCompanionDescriptor)

# ... convenience functions ...
def get_string_descriptor(string):
    """ Generates a string descriptor for the relevant string. """

    emitter = StringDescriptorEmitter()
    emitter.bString = string
    return emitter.emit()

# ... and complex emitters.

class EndpointDescriptorEmitter(ComplexDescriptorEmitter):
    """ Emitter that creates an InterfaceDescriptor. """

    DESCRIPTOR_FORMAT = EndpointDescriptor

    @contextmanager
    def SuperSpeedCompanion(self):
        """ Context manager that allows addition of a SuperSpeed Companion to this endpoint descriptor.

        It can be used with a `with` statement; and yields an SuperSpeedEndpointCompanionDescriptorEmitter
        that can be populated:

            with endpoint.SuperSpeedEndpointCompanion() as d:
                d.bMaxBurst = 1

        This adds the relevant descriptor, automatically.
        """

        descriptor = SuperSpeedEndpointCompanionDescriptorEmitter()
        yield descriptor

        self.add_subordinate_descriptor(descriptor)


class InterfaceDescriptorEmitter(ComplexDescriptorEmitter):
    """ Emitter that creates an InterfaceDescriptor. """

    DESCRIPTOR_FORMAT = InterfaceDescriptor

    @contextmanager
    def EndpointDescriptor(self, *, add_default_superspeed=False):
        """ Context manager that allows addition of a subordinate endpoint descriptor.

        It can be used with a `with` statement; and yields an EndpointDesriptorEmitter
        that can be populated:

            with interface.EndpointDescriptor() as d:
                d.bEndpointAddress = 0x01
                d.bmAttributes     = 0x80
                d.wMaxPacketSize   = 64
                d.bInterval        = 0

        This adds the relevant descriptor, automatically.
        """

        descriptor = EndpointDescriptorEmitter()
        yield descriptor

        # If we're adding a default SuperSpeed extension, do so.
        if add_default_superspeed:
            with descriptor.SuperSpeedCompanion():
                pass

        self.add_subordinate_descriptor(descriptor)


    def _pre_emit(self):

        # Count our endpoints, and update our internal count.
        self.bNumEndpoints = self._type_counts[StandardDescriptorNumbers.ENDPOINT]

        # Ensure that our interface string is an index, if we can.
        if self._collection and hasattr(self, 'iInterface'):
            self.iInterface = self._collection.ensure_string_field_is_index(self.iInterface)



class ConfigurationDescriptorEmitter(ComplexDescriptorEmitter):
    """ Emitter that creates a configuration descriptor. """

    DESCRIPTOR_FORMAT = ConfigurationDescriptor

    @contextmanager
    def InterfaceDescriptor(self):
        """ Context manager that allows addition of a subordinate interface descriptor.

        It can be used with a `with` statement; and yields an InterfaceDescriptorEmitter
        that can be populated:

            with interface.InterfaceDescriptor() as d:
                d.bInterfaceNumber = 0x01
                [snip]

        This adds the relevant descriptor, automatically. Note that populating derived
        fields such as bNumEndpoints aren't necessary; they'll be populated automatically.
        """
        descriptor = InterfaceDescriptorEmitter(collection=self._collection)
        yield descriptor

        self.add_subordinate_descriptor(descriptor)


    def _pre_emit(self):

        # Count our interfaces. Alternate settings of the same interface do not count multiple times.
        self.bNumInterfaces = len(set([subordinate[2] for subordinate in self._subordinates if (subordinate[1] == StandardDescriptorNumbers.INTERFACE)]))

        # Figure out our total length.
        subordinate_length = sum(len(sub) for sub in self._subordinates)
        self.wTotalLength = subordinate_length + self.DESCRIPTOR_FORMAT.sizeof()

        # Ensure that our configuration string is an index, if we can.
        if self._collection and hasattr(self, 'iConfiguration'):
            self.iConfiguration = self._collection.ensure_string_field_is_index(self.iConfiguration)



class DeviceDescriptorCollection:
    """ Object that builds a full collection of descriptors related to a given USB device. """

    # Most systems seem happiest with en_US (ugh), so default to that.
    DEFAULT_SUPPORTED_LANGUAGES = [LanguageIDs.ENGLISH_US]


    def __init__(self, automatic_language_descriptor=True):
        """
        Parameters:
            automatic_language_descriptor -- If set or not provided, a language descriptor will automatically
                                             be added if none exists.
        """


        self._automatic_language_descriptor = automatic_language_descriptor

        # Create our internal descriptor tracker.
        # Keys are a tuple of (type, index).
        self._descriptors = {}

        # Track string descriptors as they're created.
        self._next_string_index = 1
        self._index_for_string = {}


    def ensure_string_field_is_index(self, field_value):
        """ Processes the given field value; if it's not an string index, converts it to one.

        Non-index-fields are converted to indices using `get_index_for_string`, which automatically
        adds the relevant fields to our string descriptor collection.
        """

        if isinstance(field_value, str):
            return self.get_index_for_string(field_value)
        else:
            return field_value


    def get_index_for_string(self, string):
        """ Returns an string descriptor index for the given string.

        If a string descriptor already exists for the given string, its index is
        returned. Otherwise, a string descriptor is created.
        """

        # If we already have a descriptor for this string, return it.
        if string in self._index_for_string:
            return self._index_for_string[string]


        # Otherwise, create one:

        # Allocate an index...
        index = self._next_string_index
        self._index_for_string[string] = index
        self._next_string_index += 1

        # ... store our string descriptor with it ...
        identifier = StandardDescriptorNumbers.STRING, index
        self._descriptors[identifier] = get_string_descriptor(string)

        # ... and return our index.
        return index


    def add_descriptor(self, descriptor, index=0, descriptor_type=None):
        """ Adds a descriptor to our collection.

        Parameters:
            descriptor      -- The descriptor to be added.
            index           -- The index of the relevant descriptor. Defaults to 0.
            descriptor_type -- The type of the descriptor to be added. If `None`,
                               this is automatically derived from the descriptor contents.
        """

        # If this is an emitter rather than a descriptor itself, convert it.
        if hasattr(descriptor, 'emit'):
            descriptor = descriptor.emit()

        # Figure out the identifier (type + index) for this descriptor...
        if (descriptor_type is None):
            descriptor_type = descriptor[1]
        identifier = descriptor_type, index

        # ... and store it.
        self._descriptors[identifier] = descriptor


    def add_language_descriptor(self, supported_languages=None):
        """ Adds a language descriptor to the list of device descriptors.

        Parameters:
            supported_languages -- A list of languages supported by the device.
        """

        if supported_languages is None:
            supported_languages = self.DEFAULT_SUPPORTED_LANGUAGES

        descriptor = StringLanguageDescriptorEmitter()
        descriptor.wLANGID = supported_languages
        self.add_descriptor(descriptor)


    @contextmanager
    def DeviceDescriptor(self):
        """ Context manager that allows addition of a device descriptor.

        It can be used with a `with` statement; and yields an DeviceDescriptorEmitter
        that can be populated:

            with collection.DeviceDescriptor() as d:
                d.idVendor  = 0xabcd
                d.idProduct = 0x1234
                [snip]

        This adds the relevant descriptor, automatically.
        """
        descriptor = DeviceDescriptorEmitter()
        yield descriptor

        # If we have any string fields, ensure that they're indices before continuing.
        for field in ('iManufacturer', 'iProduct', 'iSerialNumber'):
            if hasattr(descriptor, field):
                value = getattr(descriptor, field)
                index = self.ensure_string_field_is_index(value)
                setattr(descriptor, field, index)

        self.add_descriptor(descriptor)


    @contextmanager
    def ConfigurationDescriptor(self):
        """ Context manager that allows addition of a configuration descriptor.

        It can be used with a `with` statement; and yields an ConfigurationDescriptorEmitter
        that can be populated:

            with collection.ConfigurationDescriptor() as d:
                d.bConfigurationValue = 1
                [snip]

        This adds the relevant descriptor, automatically. Note that populating derived
        fields such as bNumInterfaces aren't necessary; they'll be populated automatically.
        """
        descriptor = ConfigurationDescriptorEmitter(collection=self)
        yield descriptor

        self.add_descriptor(descriptor)


    def _ensure_has_language_descriptor(self):
        """ ensures that we have a language descriptor; adding one if necessary."""

        # if we're not automatically adding a language descriptor, we shouldn't do anything,
        # and we'll just ignore this.
        if not self._automatic_language_descriptor:
            return

        # if we don't have a language descriptor, add our default one.
        if (StandardDescriptorNumbers.STRING, 0) not in self._descriptors:
            self.add_language_descriptor()



    def get_descriptor_bytes(self, type_number: int, index: int = 0):
        """ Returns the raw, binary descriptor for a given descriptor type/index.

        Parmeters:
            type_number -- The descriptor type number.
            index       -- The index of the relevant descriptor, if relevant.
        """

        # If this is a request for a language descriptor, return one.
        if (type_number, index) == (StandardDescriptorNumbers.STRING, 0):
            self._ensure_has_language_descriptor()

        return self._descriptors[(type_number, index)]


    def __iter__(self):
        """ Allow iterating over each of our descriptors; yields (index, value, descriptor). """
        self._ensure_has_language_descriptor()
        return ((number, index, desc) for ((number, index), desc) in self._descriptors.items())




class BinaryObjectStoreDescriptorEmitter(ComplexDescriptorEmitter):
    """ Emitter that creates a BinaryObjectStore descriptor. """

    DESCRIPTOR_FORMAT = BinaryObjectStoreDescriptor

    @contextmanager
    def USB2Extension(self):
        """ Context manager that allows addition of a USB 2.0 Extension to this Binary Object Store.

        It can be used with a `with` statement; and yields an USB2ExtensionDescriptorEmitter
        that can be populated:

            with bos.USB2Extension() as e:
                e.bmAttributes = 1

        This adds the relevant descriptor, automatically.
        """

        descriptor = USB2ExtensionDescriptorEmitter()
        yield descriptor

        self.add_subordinate_descriptor(descriptor)


    @contextmanager
    def SuperSpeedUSBDeviceCapability(self):
        """ Context manager that allows addition of a SS Device Capability to this Binary Object Store.

        It can be used with a `with` statement; and yields an SuperSpeedUSBDeviceCapabilityDescriptorEmitter
        that can be populated:

            with bos.SuperSpeedUSBDeviceCapability() as e:
                e.wSpeedSupported       = 0b1110
                e.bFunctionalitySupport = 1

        This adds the relevant descriptor, automatically.
        """

        descriptor = SuperSpeedUSBDeviceCapabilityDescriptorEmitter()
        yield descriptor

        self.add_subordinate_descriptor(descriptor)


    def _pre_emit(self):

        # Figure out the total length of our descriptor, including subordinates.
        subordinate_length = sum(len(sub) for sub in self._subordinates)
        self.wTotalLength = subordinate_length + self.DESCRIPTOR_FORMAT.sizeof()

        # Count our subordinate descriptors, and update our internal count.
        self.bNumDeviceCaps = len(self._subordinates)



class SuperSpeedDeviceDescriptorCollection(DeviceDescriptorCollection):
    """ Object that builds a full collection of descriptors related to a given USB3 device. """

    def __init__(self, automatic_descriptors=True):
        """
        Parameters:
            automatic_descriptors -- If set or not provided, certian required descriptors will be
                                     be added if none exists.
        """
        self._automatic_descriptors = automatic_descriptors
        super().__init__(automatic_language_descriptor=automatic_descriptors)


    @contextmanager
    def BOSDescriptor(self):
        """ Context manager that allows addition of a Binary Object Store descriptor.

        It can be used with a `with` statement; and yields an BinaryObjectStoreDescriptorEmitter
        that can be populated:

            with collection.BOSDescriptor() as d:
                [snip]

        This adds the relevant descriptor, automatically. Note that populating derived
        fields such as bNumDeviceCaps aren't necessary; they'll be populated automatically.
        """
        descriptor = BinaryObjectStoreDescriptorEmitter()
        yield descriptor

        self.add_descriptor(descriptor)


    def add_default_bos_descriptor(self):
        """ Adds a default, empty BOS descriptor. """

        # Create an empty BOS descriptor...
        descriptor = BinaryObjectStoreDescriptorEmitter()

        # ... populate our default required descriptors...
        descriptor.add_subordinate_descriptor(USB2ExtensionDescriptorEmitter())
        descriptor.add_subordinate_descriptor(SuperSpeedUSBDeviceCapabilityDescriptorEmitter())

        # ... and add it to our overall BOS descriptor.
        self.add_descriptor(descriptor)


    def _ensure_has_bos_descriptor(self):
        """ Ensures that we have a BOS descriptor; adding one if necessary."""

        # If we're not automatically adding a language descriptor, we shouldn't do anything,
        # and we'll just ignore this.
        if not self._automatic_descriptors:
            return

        # If we don't have a language descriptor, add our default one.
        if (StandardDescriptorNumbers.BOS, 0) not in self._descriptors:
            self.add_default_bos_descriptor()


    def __iter__(self):
        """ Allow iterating over each of our descriptors; yields (index, value, descriptor). """
        self._ensure_has_bos_descriptor()
        return super().__iter__()


class EmitterTests(unittest.TestCase):

    def test_string_emitter(self):
        emitter = StringDescriptorEmitter()
        emitter.bString = "Hello"

        self.assertEqual(emitter.emit(), b"\x0C\x03H\0e\0l\0l\0o\0")


    def test_string_emitter_function(self):
        self.assertEqual(get_string_descriptor("Hello"), b"\x0C\x03H\0e\0l\0l\0o\0")


    def test_configuration_emitter(self):
        descriptor = bytes([

            # config descriptor
            12,     # length
            2,      # type
            25, 00, # total length
            1,      # num interfaces
            1,      # configuration number
            0,      # config string
            0x80,   # attributes
            250,    # max power

            # interface descriptor
            9,    # length
            4,    # type
            0,    # number
            0,    # alternate
            1,    # num endpoints
            0xff, # class
            0xff, # subclass
            0xff, # protocol
            0,    # string

            # endpoint descriptor
            7,       # length
            5,       # type
            0x01,    # address
            2,       # attributes
            64, 0,   # max packet size
            255,     # interval
        ])


        # Create a trivial configuration descriptor...
        emitter = ConfigurationDescriptorEmitter()

        with emitter.InterfaceDescriptor() as interface:
            interface.bInterfaceNumber = 0

            with interface.EndpointDescriptor() as endpoint:
                endpoint.bEndpointAddress = 1


        # ... and validate that it maches our reference descriptor.
        binary = emitter.emit()
        self.assertEqual(len(binary), len(descriptor))


    def test_descriptor_collection(self):
        collection = DeviceDescriptorCollection()

        with collection.DeviceDescriptor() as d:
            d.idVendor           = 0xdead
            d.idProduct          = 0xbeef
            d.bNumConfigurations = 1

            d.iManufacturer      = "Test Company"
            d.iProduct           = "Test Product"


        with collection.ConfigurationDescriptor() as c:
            c.bConfigurationValue = 1

            with c.InterfaceDescriptor() as i:
                i.bInterfaceNumber = 1

                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = 0x81

                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = 0x01


        results = list(collection)

        # We should wind up with four descriptor entries, as our endpoint/interface descriptors are
        # included in our configuration descriptor.
        self.assertEqual(len(results), 5)

        # Supported languages string.
        self.assertIn((3, 0, b'\x04\x03\x09\x04'), results)

        # Manufacturer / product string.
        self.assertIn((3, 1, b'\x1a\x03T\x00e\x00s\x00t\x00 \x00C\x00o\x00m\x00p\x00a\x00n\x00y\x00'), results)
        self.assertIn((3, 2, b'\x1a\x03T\x00e\x00s\x00t\x00 \x00P\x00r\x00o\x00d\x00u\x00c\x00t\x00'), results)

        # Device descriptor.
        self.assertIn((1, 0, b'\x12\x01\x00\x02\x00\x00\x00@\xad\xde\xef\xbe\x00\x00\x01\x02\x00\x01'), results)

        # Configuration descriptor, with subordinates.
        self.assertIn((2, 0, b'\t\x02 \x00\x01\x01\x00\x80\xfa\t\x04\x01\x00\x02\xff\xff\xff\x00\x07\x05\x81\x02@\x00\xff\x07\x05\x01\x02@\x00\xff'), results)


    def test_empty_descriptor_collection(self):
        collection = DeviceDescriptorCollection(automatic_language_descriptor=False)
        results = list(collection)
        self.assertEqual(len(results), 0)

    def test_automatic_language_descriptor(self):
        collection = DeviceDescriptorCollection(automatic_language_descriptor=True)
        results = list(collection)
        self.assertEqual(len(results), 1)

if __name__ == "__main__":
    unittest.main()



#
# This file is part of usb-protocol.
#
""" USB types -- defines enumerations that describe standard USB types """

from enum import Enum, IntFlag, IntEnum

class USBDirection(IntEnum):
    """ Class representing USB directions. """
    OUT = 0
    IN = 1

    def is_in(self):
        return self is self.IN

    def is_out(self):
        return self is self.OUT

    @classmethod
    def parse(cls, value):
        """ Helper that converts a numeric field into a direction. """
        return cls(value)

    @classmethod
    def from_request_type(cls, request_type_int):
        """ Helper method that extracts the direction from a request_type integer. """
        return cls(request_type_int >> 7)

    @classmethod
    def from_endpoint_address(cls, address):
        """ Helper method that extracts the direction from an endpoint address. """
        return cls(address >> 7)

    def token(self):
        """ Generates the token corresponding to the given direction. """
        return USBPacketID.IN if (self is self.IN) else USBPacketID.OUT

    def reverse(self):
        """ Returns the reverse of the given direction. """
        return self.OUT if (self is self.IN) else self.IN


    def to_endpoint_address(self, endpoint_number):
        """ Helper method that converts and endpoint_number to an address, given direction. """
        if self.is_in():
            return endpoint_number | (1 << 7)
        else:
            return endpoint_number


class USBPIDCategory(IntFlag):
    """ Category constants for each of the groups that PIDs can fall under. """

    SPECIAL   = 0b00
    TOKEN     = 0b01
    HANDSHAKE = 0b10
    DATA      = 0b11

    MASK      = 0b11



class USBPacketID(IntFlag):
    """ Enumeration specifying all of the valid USB PIDs we can handle. """

    # Token group (lsbs = 0b01).
    OUT   = 0b0001
    IN    = 0b1001
    SOF   = 0b0101
    SETUP = 0b1101

    # Data group (lsbs = 0b11).
    DATA0 = 0b0011
    DATA1 = 0b1011
    DATA2 = 0b0111
    MDATA = 0b1111

    # Handshake group (lsbs = 0b10)
    ACK   = 0b0010
    NAK   = 0b1010
    STALL = 0b1110
    NYET  = 0b0110

    # Special group.
    PRE   = 0b1100
    ERR   = 0b1100
    SPLIT = 0b1000
    PING  = 0b0100

    # Flag representing that the PID seems invalid.
    PID_INVALID   = 0b10000
    PID_CORE_MASK = 0b01111


    @classmethod
    def from_byte(cls, byte, skip_checks=False):
        """ Creates a PID object from a byte. """

        # Convert the raw PID to an integer.
        pid_as_int = int.from_bytes(byte, byteorder='little')
        return cls.from_int(pid_as_int, skip_checks=skip_checks)


    @classmethod
    def from_int(cls, value, skip_checks=True):
        """ Create a PID object from an integer. """

        PID_MASK           = 0b1111
        INVERTED_PID_SHIFT = 4

        # Pull out the PID and its inverse from the byte.
        pid          = cls(value & PID_MASK)
        inverted_pid = value >> INVERTED_PID_SHIFT

        # If we're not skipping checks,
        if not skip_checks:
            if (pid ^ inverted_pid) != PID_MASK:
                pid |= cls.PID_INVALID

        return cls(pid)


    @classmethod
    def from_name(cls, name):
        """ Create a PID object from a string representation of its name. """
        return cls[name]


    @classmethod
    def parse(cls, value):
        """ Attempt to create a PID object from a number, byte, or string. """

        if isinstance(value, bytes):
            return cls.from_byte(value)

        if isinstance(value, str):
            return cls.from_name(value)

        if isinstance(value, int):
            return cls.from_int(value)

        return cls(value)


    def category(self):
        """ Returns the USBPIDCategory that each given PID belongs to. """
        return USBPIDCategory(self & USBPIDCategory.MASK)


    def is_data(self):
        """ Returns true iff the given PID represents a DATA packet. """
        return self.category() is USBPIDCategory.DATA


    def is_token(self):
        """ Returns true iff the given PID represents a token packet. """
        return self.category() is USBPIDCategory.TOKEN


    def is_handshake(self):
        """ Returns true iff the given PID represents a handshake packet. """
        return self.category() is USBPIDCategory.HANDSHAKE


    def is_invalid(self):
        """ Returns true if this object is an attempt to encapsulate an invalid PID. """
        return (self & self.PID_INVALID)

    def direction(self):
        """ Get a USB direction from a PacketID. """

        if self is self.SOF:
            return None

        if self is self.SETUP or self is self.OUT:
            return USBDirection.OUT

        if self is self.IN:
            return USBDirection.IN

        raise ValueError("cannot determine the direction of a non-token PID")


    def summarize(self):
        """ Return a summary of the given packet. """

        # By default, get the raw name.
        core_pid  = self & self.PID_CORE_MASK
        name = core_pid.name

        if self.is_invalid():
            return "{} (check-nibble invalid)".format(name)
        else:
            return name


    def byte(self):
        """ Return the PID's value with its upper nibble. """

        inverted_pid = self ^ 0b1111
        full_pid     = (inverted_pid << 4) | self

        return full_pid


class USBRequestRecipient(IntEnum):
    """ Enumeration that describes each 'recipient' of a USB request field. """

    DEVICE    = 0
    INTERFACE = 1
    ENDPOINT  = 2
    OTHER     = 3

    RESERVED  = 4

    @classmethod
    def from_integer(cls, value):
        """ Special factory that correctly handles reserved values. """

        # If we have one of the reserved values; indicate so.
        if 4 <= value < 16:
            return cls.RESERVED

        # Otherwise, translate the raw value.
        return cls(value)


    @classmethod
    def from_request_type(cls, request_type_int):
        """ Helper method that extracts the type from a request_type integer. """

        MASK  = 0b11111
        return cls(request_type_int & MASK)


class USBRequestType(IntEnum):
    """ Enumeration that describes each possible Type field for a USB request. """

    STANDARD  = 0
    CLASS     = 1
    VENDOR    = 2
    RESERVED  = 3


    @classmethod
    def from_request_type(cls, request_type_int):

        """ Helper method that extracts the type from a request_type integer. """
        SHIFT = 5
        MASK  = 0b11

        return cls((request_type_int >> SHIFT) & MASK)



class USBTransferType(IntEnum):
    CONTROL     = 0
    ISOCHRONOUS = 1
    BULK        = 2
    INTERRUPT   = 3


def endpoint_number_from_address(number):
    return number & 0x7F


LANGUAGE_NAMES = {
    0x0436: "Afrikaans",
    0x041c: "Albanian",
    0x0401: "Arabic (Saudi Arabia)",
    0x0801: "Arabic (Iraq)",
    0x0c01: "Arabic (Egypt)",
    0x1001: "Arabic (Libya)",
    0x1401: "Arabic (Algeria)",
    0x1801: "Arabic (Morocco)",
    0x1c01: "Arabic (Tunisia)",
    0x2001: "Arabic (Oman)",
    0x2401: "Arabic (Yemen)",
    0x2801: "Arabic (Syria)",
    0x2c01: "Arabic (Jordan)",
    0x3001: "Arabic (Lebanon)",
    0x3401: "Arabic (Kuwait)",
    0x3801: "Arabic (U.A.E.)",
    0x3c01: "Arabic (Bahrain)",
    0x4001: "Arabic (Qatar)",
    0x042b: "Armenian",
    0x044d: "Assamese",
    0x042c: "Azeri (Latin)",
    0x082c: "Azeri (Cyrillic)",
    0x042d: "Basque",
    0x0423: "Belarussian",
    0x0445: "Bengali",
    0x0402: "Bulgarian",
    0x0455: "Burmese",
    0x0403: "Catalan",
    0x0404: "Chinese (Taiwan)",
    0x0804: "Chinese (PRC)",
    0x0c04: "Chinese (Hong Kong SAR, PRC)",
    0x1004: "Chinese (Singapore)",
    0x1404: "Chinese (Macau SAR)",
    0x041a: "Croatian",
    0x0405: "Czech",
    0x0406: "Danish",
    0x0413: "Dutch (Netherlands)",
    0x0813: "Dutch (Belgium)",
    0x0409: "English (US)",
    0x0809: "English (United Kingdom)",
    0x0c09: "English (Australian)",
    0x1009: "English (Canadian)",
    0x1409: "English (New Zealand)",
    0x1809: "English (Ireland)",
    0x1c09: "English (South Africa)",
    0x2009: "English (Jamaica)",
    0x2409: "English (Caribbean)",
    0x2809: "English (Belize)",
    0x2c09: "English (Trinidad)",
    0x3009: "English (Zimbabwe)",
    0x3409: "English (Philippines)",
    0x0425: "Estonian",
    0x0438: "Faeroese",
    0x0429: "Farsi",
    0x040b: "Finnish",
    0x040c: "French (Standard)",
    0x080c: "French (Belgian)",
    0x0c0c: "French (Canadian)",
    0x100c: "French (Switzerland)",
    0x140c: "French (Luxembourg)",
    0x180c: "French (Monaco)",
    0x0437: "Georgian",
    0x0407: "German (Standard)",
    0x0807: "German (Switzerland)",
    0x0c07: "German (Austria)",
    0x1007: "German (Luxembourg)",
    0x1407: "German (Liechtenstein)",
    0x0408: "Greek",
    0x0447: "Gujarati",
    0x040d: "Hebrew",
    0x0439: "Hindi",
    0x040e: "Hungarian",
    0x040f: "Icelandic",
    0x0421: "Indonesian",
    0x0410: "Italian (Standard)",
    0x0810: "Italian (Switzerland)",
    0x0411: "Japanese",
    0x044b: "Kannada",
    0x0860: "Kashmiri (India)",
    0x043f: "Kazakh",
    0x0457: "Konkani",
    0x0412: "Korean",
    0x0812: "Korean (Johab)",
    0x0426: "Latvian",
    0x0427: "Lithuanian",
    0x0827: "Lithuanian (Classic)",
    0x042f: "Macedonian",
    0x043e: "Malay (Malaysian)",
    0x083e: "Malay (Brunei Darussalam)",
    0x044c: "Malayalam",
    0x0458: "Manipuri",
    0x044e: "Marathi",
    0x0861: "Nepali (India)",
    0x0414: "Norwegian (Bokmal)",
    0x0814: "Norwegian (Nynorsk)",
    0x0448: "Oriya",
    0x0415: "Polish",
    0x0416: "Portuguese (Brazil)",
    0x0816: "Portuguese (Standard)",
    0x0446: "Punjabi",
    0x0418: "Romanian",
    0x0419: "Russian",
    0x044f: "Sanskrit",
    0x0c1a: "Serbian (Cyrillic)",
    0x081a: "Serbian (Latin)",
    0x0459: "Sindhi",
    0x041b: "Slovak",
    0x0424: "Slovenian",
    0x040a: "Spanish (Traditional Sort)",
    0x080a: "Spanish (Mexican)",
    0x0c0a: "Spanish (Modern Sort)",
    0x100a: "Spanish (Guatemala)",
    0x140a: "Spanish (Costa Rica)",
    0x180a: "Spanish (Panama)",
    0x1c0a: "Spanish (Dominican Republic)",
    0x200a: "Spanish (Venezuela)",
    0x240a: "Spanish (Colombia)",
    0x280a: "Spanish (Peru)",
    0x2c0a: "Spanish (Argentina)",
    0x300a: "Spanish (Ecuador)",
    0x340a: "Spanish (Chile)",
    0x380a: "Spanish (Uruguay)",
    0x3c0a: "Spanish (Paraguay)",
    0x400a: "Spanish (Bolivia)",
    0x440a: "Spanish (El Salvador)",
    0x480a: "Spanish (Honduras)",
    0x4c0a: "Spanish (Nicaragua)",
    0x500a: "Spanish (Puerto Rico)",
    0x0430: "Sutu",
    0x0441: "Swahili (Kenya)",
    0x041d: "Swedish",
    0x081d: "Swedish (Finland)",
    0x0449: "Tamil",
    0x0444: "Tatar (Tatarstan)",
    0x044a: "Telugu",
    0x041e: "Thai",
    0x041f: "Turkish",
    0x0422: "Ukrainian",
    0x0420: "Urdu (Pakistan)",
    0x0820: "Urdu (India)",
    0x0443: "Uzbek (Latin)",
    0x0843: "Uzbek (Cyrillic)",
    0x042a: "Vietnamese",
    0x04ff: "HID (Usage Data Descriptor)",
    0xf0ff: "HID (Vendor Defined 1)",
    0xf4ff: "HID (Vendor Defined 2)",
    0xf8ff: "HID (Vendor Defined 3)",
    0xfcff: "HID (Vendor Defined 4)",
}


class LanguageIDs(IntEnum):
    AFRIKAANS                  = 0X0436
    ALBANIAN                   = 0X041C
    ARABIC_SAUDI_ARABIA        = 0X0401
    ARABIC_IRAQ                = 0X0801
    ARABIC_EGYPT               = 0X0C01
    ARABIC_LIBYA               = 0X1001
    ARABIC_ALGERIA             = 0X1401
    ARABIC_MOROCCO             = 0X1801
    ARABIC_TUNISIA             = 0X1C01
    ARABIC_OMAN                = 0X2001
    ARABIC_YEMEN               = 0X2401
    ARABIC_SYRIA               = 0X2801
    ARABIC_JORDAN              = 0X2C01
    ARABIC_LEBANON             = 0X3001
    ARABIC_KUWAIT              = 0X3401
    ARABIC_UAE                 = 0X3801
    ARABIC_BAHRAIN             = 0X3C01
    ARABIC_QATAR               = 0X4001
    ARMENIAN                   = 0X042B
    ASSAMESE                   = 0X044D
    AZERI_LATIN                = 0X042C
    AZERI_CYRILLIC             = 0X082C
    BASQUE                     = 0X042D
    BELARUSSIAN                = 0X0423
    BENGALI                    = 0X0445
    BULGARIAN                  = 0X0402
    BURMESE                    = 0X0455
    CATALAN                    = 0X0403
    CHINESE_TAIWAN             = 0X0404
    CHINESE_PRC                = 0X0804
    CHINESE_HONG_KONG          = 0X0C04
    CHINESE_SINGAPORE          = 0X1004
    CHINESE_MACAU_SAR          = 0X1404
    CROATIAN                   = 0X041A
    CZECH                      = 0X0405
    DANISH                     = 0X0406
    DUTCH_NETHERLANDS          = 0X0413
    DUTCH_BELGIUM              = 0X0813
    ENGLISH_US                 = 0X0409
    ENGLISH_UNITED_KINGDOM     = 0X0809
    ENGLISH_AUSTRALIAN         = 0X0C09
    ENGLISH_CANADIAN           = 0X1009
    ENGLISH_NEW_ZEALAND        = 0X1409
    ENGLISH_IRELAND            = 0X1809
    ENGLISH_SOUTH_AFRICA       = 0X1C09
    ENGLISH_JAMAICA            = 0X2009
    ENGLISH_CARIBBEAN          = 0X2409
    ENGLISH_BELIZE             = 0X2809
    ENGLISH_TRINIDAD           = 0X2C09
    ENGLISH_ZIMBABWE           = 0X3009
    ENGLISH_PHILIPPINES        = 0X3409
    ESTONIAN                   = 0X0425
    FAEROESE                   = 0X0438
    FARSI                      = 0X0429
    FINNISH                    = 0X040B
    FRENCH_STANDARD            = 0X040C
    FRENCH_BELGIAN             = 0X080C
    FRENCH_CANADIAN            = 0X0C0C
    FRENCH_SWITZERLAND         = 0X100C
    FRENCH_LUXEMBOURG          = 0X140C
    FRENCH_MONACO              = 0X180C
    GEORGIAN                   = 0X0437
    GERMAN_STANDARD            = 0X0407
    GERMAN_SWITZERLAND         = 0X0807
    GERMAN_AUSTRIA             = 0X0C07
    GERMAN_LUXEMBOURG          = 0X1007
    GERMAN_LIECHTENSTEIN       = 0X1407
    GREEK                      = 0X0408
    GUJARATI                   = 0X0447
    HEBREW                     = 0X040D
    HINDI                      = 0X0439
    HUNGARIAN                  = 0X040E
    ICELANDIC                  = 0X040F
    INDONESIAN                 = 0X0421
    ITALIAN_STANDARD           = 0X0410
    ITALIAN_SWITZERLAND        = 0X0810
    JAPANESE                   = 0X0411
    KANNADA                    = 0X044B
    KASHMIRI_INDIA             = 0X0860
    KAZAKH                     = 0X043F
    KONKANI                    = 0X0457
    KOREAN                     = 0X0412
    KOREAN_JOHAB               = 0X0812
    LATVIAN                    = 0X0426
    LITHUANIAN                 = 0X0427
    LITHUANIAN_CLASSIC         = 0X0827
    MACEDONIAN                 = 0X042F
    MALAY_MALAYSIAN            = 0X043E
    MALAY_BRUNEI_DARUSSALAM    = 0X083E
    MALAYALAM                  = 0X044C
    MANIPURI                   = 0X0458
    MARATHI                    = 0X044E
    NEPALI_INDIA               = 0X0861
    NORWEGIAN_BOKMAL           = 0X0414
    NORWEGIAN_NYNORSK          = 0X0814
    ORIYA                      = 0X0448
    POLISH                     = 0X0415
    PORTUGUESE_BRAZIL          = 0X0416
    PORTUGUESE_STANDARD        = 0X0816
    PUNJABI                    = 0X0446
    ROMANIAN                   = 0X0418
    RUSSIAN                    = 0X0419
    SANSKRIT                   = 0X044F
    SERBIAN_CYRILLIC           = 0X0C1A
    SERBIAN_LATIN              = 0X081A
    SINDHI                     = 0X0459
    SLOVAK                     = 0X041B
    SLOVENIAN                  = 0X0424
    SPANISH_TRADITIONAL_SORT   = 0X040A
    SPANISH_MEXICAN            = 0X080A
    SPANISH_MODERN_SORT        = 0X0C0A
    SPANISH_GUATEMALA          = 0X100A
    SPANISH_COSTA_RICA         = 0X140A
    SPANISH_PANAMA             = 0X180A
    SPANISH_DOMINICAN_REPUBLIC = 0X1C0A
    SPANISH_VENEZUELA          = 0X200A
    SPANISH_COLOMBIA           = 0X240A
    SPANISH_PERU               = 0X280A
    SPANISH_ARGENTINA          = 0X2C0A
    SPANISH_ECUADOR            = 0X300A
    SPANISH_CHILE              = 0X340A
    SPANISH_URUGUAY            = 0X380A
    SPANISH_PARAGUAY           = 0X3C0A
    SPANISH_BOLIVIA            = 0X400A
    SPANISH_EL_SALVADOR        = 0X440A
    SPANISH_HONDURAS           = 0X480A
    SPANISH_NICARAGUA          = 0X4C0A
    SPANISH_PUERTO_RICO        = 0X500A
    SUTU                       = 0X0430
    SWAHILI_KENYA              = 0X0441
    SWEDISH                    = 0X041D
    SWEDISH_FINLAND            = 0X081D
    TAMIL                      = 0X0449
    TATAR_TATARSTAN            = 0X0444
    TELUGU                     = 0X044A
    THAI                       = 0X041E
    TURKISH                    = 0X041F
    UKRAINIAN                  = 0X0422
    URDU_PAKISTAN              = 0X0420
    URDU_INDIA                 = 0X0820
    UZBEK_LATIN                = 0X0443
    UZBEK_CYRILLIC             = 0X0843
    VIETNAMESE                 = 0X042A
    HID_USAGE_DATA_DESCRIPTOR  = 0X04FF
    HID_VENDOR_DEFINED_1       = 0XF0FF
    HID_VENDOR_DEFINED_2       = 0XF4FF
    HID_VENDOR_DEFINED_3       = 0XF8FF
    HID_VENDOR_DEFINED_4       = 0XFCFF


class DescriptorTypes(IntEnum):
    DEVICE                    = 1
    CONFIGURATION             = 2
    STRING                    = 3
    INTERFACE                 = 4
    ENDPOINT                  = 5
    DEVICE_QUALIFIER          = 6
    OTHER_SPEED_CONFIGURATION = 7
    INTERFACE_POWER           = 8
    HID                       = 33
    REPORT                    = 34


class USBSynchronizationType(IntEnum):
    NONE         = 0x00
    ASYNC        = 0x01
    ADAPTIVE     = 0x02
    SYNCHRONOUS  = 0x03


class USBUsageType(IntEnum):
    DATA              = 0
    FEEDBACK          = 1
    IMPLICIT_FEEDBACK = 2


class USBTransferType(IntEnum):
    CONTROL     = 0
    ISOCHRONOUS = 1
    BULK        = 2
    INTERRUPT   = 3


class USBSynchronizationType(IntEnum):
    NONE         = 0x00
    ASYNC        = 0x01
    ADAPTIVE     = 0x02
    SYNCHRONOUS  = 0x03


class USBUsageType(IntEnum):
    DATA              = 0
    FEEDBACK          = 1
    IMPLICIT_FEEDBACK = 2


class USBStandardRequests(IntEnum):
    GET_STATUS        = 0
    CLEAR_FEATURE     = 1
    SET_FEATURE       = 3
    SET_ADDRESS       = 5
    GET_DESCRIPTOR    = 6
    SET_DESCRIPTOR    = 7
    GET_CONFIGURATION = 8
    SET_CONFIGURATION = 9
    GET_INTERFACE     = 10
    SET_INTERFACE     = 11
    SYNCH_FRAME       = 12

    # USB3 only.
    SET_ENCRYPTION       = 13
    GET_ENCRYPTION       = 14
    SET_HANDSHAKE        = 15
    GET_HANDSHAKE        = 16
    SET_CONNECTION       = 17
    SET_SECURITY_DATA    = 18
    GET_SECURITY_DATA    = 19
    SET_WUSB_DATA        = 20
    LOOPBACK_DATA_WRITE  = 21
    LOOPBACK_DATA_READ   = 22
    SET_INTERFACE_DS     = 23
    SET_SEL              = 48
    SET_ISOCH_DELAY      = 49






#
# This file is part of usb-protocol.
#
""" Type elements for defining USB descriptors. """

import unittest
import construct

class DescriptorFormat(construct.Struct):
    """
    Creates a Construct structure for a USB descriptor, and a corresponding version that
    supports parsing incomplete binary as `DescriptorType.Partial`, e.g. `DeviceDescriptor.Partial`.
    """

    def __init__(self, *subcons, _create_partial=True, **subconskw):

        if _create_partial:
            self.Partial = self._create_partial(*subcons, **subconskw) # pylint: disable=invalid-name

        super().__init__(*subcons, **subconskw)


    @classmethod
    def _get_subcon_field_type(cls, subcon):
        """ Gets the actual field type for a Subconstruct behind arbitrary levels of `Renamed`s."""

        # DescriptorFields are usually `<Renamed <Renamed <FormatField>>>`.
        # The type behind the `Renamed`s is the one we're interested in, so let's recursively examine
        # the child Subconstruct until we get to it.

        if not isinstance(subcon, construct.Renamed):
            return subcon
        else:
            return cls._get_subcon_field_type(subcon.subcon)


    @classmethod
    def _create_partial(cls, *subcons, **subconskw):
        """ Creates a version of the descriptor format for parsing incomplete binary data as a descriptor.

        This essentially wraps every field after bLength and bDescriptorType in a `construct.Optional`.
        """

        def _apply_optional(subcon):

            subcon_type = cls._get_subcon_field_type(subcon)

            #
            # If it's already Optional then we don't need to apply it again.
            #
            if isinstance(subcon_type, construct.Select):
                # construct uses a weird singleton to define Pass. `construct.core.Pass` would normally be
                # the type's name, but then they create a singleton of that same name, replacing that name and
                # making the type technically unnamable and only accessable via `type()`.
                if isinstance(subcon_type.subcons[1], type(construct.Pass)):
                    return subcon

            return (subcon.name / construct.Optional(subcon_type)) * subcon.docs

        # First store the Subconstructs we don't want to modify: bLength and bDescriptorType,
        # as these are never optional.
        new_subcons = list(subcons[0:2])

        # Then apply Optional to all of the rest of the Subconstructs.
        new_subcons.extend([_apply_optional(subcon) for subcon in subcons[2:]])

        return DescriptorFormat(*new_subcons, _create_partial=False, **subconskw)


    @staticmethod
    def _to_detail_dictionary(descriptor, use_pretty_names=True):
        result = {}

        # Loop over every entry in our descriptor context, and try to get a
        # fancy name for it.
        for key, value in descriptor.items():

            # Don't include any underscore-prefixed private members.
            if key.startswith('_'):
                continue

            # If there's no definition for the given key in our format, # skip it.
            if not hasattr(descriptor._format, key):
                continue

            # Try to apply any documentation on the given field rather than it's internal name.
            format_element = getattr(descriptor._format, key)
            detail_key = format_element.docs if (format_element.docs and use_pretty_names) else key

            # Finally, add the entry to our dict.
            result[detail_key] = value

        return result


    def parse(self, data, **context_keywords):
        """ Hook on the parent parse() method which attaches a few methods. """

        # Use construct to run the parse itself...
        result = super().parse(bytes(data), **context_keywords)

        # ... and then bind our static to_detail_dictionary to it.
        result._format = self
        result._to_detail_dictionary = self._to_detail_dictionary.__get__(result, type(result))

        return result


class DescriptorNumber(construct.Const):
    """ Trivial wrapper class that denotes a particular Const as the descriptor number. """

    def __init__(self, const):

        # If our descriptor number is an integer, instead of "raw",
        # convert it to a byte, first.
        if not isinstance(const, bytes):
            const = const.to_bytes(1, byteorder='little')

        # Grab the inner descriptor number represented by the constant.
        self.number = int.from_bytes(const, byteorder='little')

        # And pass this to the core constant class.
        super().__init__(const)

        # Finally, add a documentation string for the type.
        self.docs = "Descriptor type"


    def _parse(self, stream, context, path):
        const_bytes = super()._parse(stream, context, path)
        return const_bytes[0]


    def get_descriptor_number(self):
        """ Returns this constant's associated descriptor number."""
        return self.number


class BCDFieldAdapter(construct.Adapter):
    """ Construct adapter that dynamically parses BCD fields. """

    def _decode(self, obj, context, path):
        hex_string = f"{obj:04x}"
        return float(f"{hex_string[0:2]}.{hex_string[2:]}")


    def _encode(self, obj, context, path):

        # Break the object down into its component parts...
        integer = int(obj) % 100
        percent = int(round(obj * 100)) % 100

        # ... make sure nothing is lost during conversion...
        if float(f"{integer:02}.{percent:02}") != obj:
            raise AssertionError("BCD fields must be in the format XX.YY")

        # ... and squish them into an integer.
        return int(f"{integer:02}{percent:02}", 16)



class DescriptorField(construct.Subconstruct):
    """
    Construct field definition that automatically adds fields of the proper
    size to Descriptor definitions.
    """

    #
    # The C++-wonk operator overloading is Construct, not me, I swear.
    #

    # FIXME: these are really primitive views of these types;
    # we should extend these to get implicit parsing wherever possible
    USB_TYPES = {
        'b'   : construct.Int8ul,
        'bcd' : BCDFieldAdapter(construct.Int16ul),  # TODO: Create a BCD parser for this
        'i'   : construct.Int8ul,
        'id'  : construct.Int16ul,
        'bm'  : construct.Int8ul,
        'w'   : construct.Int16ul,
    }


    LENGTH_TYPES = {
        1: construct.Int8ul,
        2: construct.Int16ul,
        3: construct.Int24ul,
        4: construct.Int32ul,
        8: construct.Int64ul
    }


    @staticmethod
    def _get_prefix(name):
        """ Returns the lower-case prefix on a USB descriptor name. """
        prefix = []

        # Silly loop that continues until we find an uppercase letter.
        # You'd be aghast at how the 'pythonic' answers look.
        for c in name:

            # Ignore leading underscores.
            if c == '_':
                continue

            if c.isupper():
                break
            prefix.append(c)

        return ''.join(prefix)


    @classmethod
    def _get_type_for_name(cls, name):
        """ Returns the type that's appropriate for a given descriptor field name. """

        try:
            return cls.USB_TYPES[cls._get_prefix(name)]
        except KeyError:
            raise ValueError("field names must be formatted per the USB standard!")


    def __init__(self, description="", default=None, *, length=None):
        self.description = description
        self.default = default
        self.length = length


    def __rtruediv__(self, field_name):
        # If we have a length, use it to figure out the type.
        # Otherwise, extract the type from the prefix. (Using a length
        # is useful for e.g. USB3 bitfields; which can span several bytes.)
        if self.length is not None:
            field_type = self.LENGTH_TYPES[self.length]
        else:
            field_type = self._get_type_for_name(field_name)

        if self.default is not None:
            field_type = construct.Default(field_type, self.default)

        # Build our subconstruct. Construct makes this look super weird,
        # but this is actually "we have a field with <field_name> of type <field_type>".
        # In long form, we'll call it "description".
        return (field_name / field_type) * self.description


# Convenience type that gets a descriptor's own length.
DescriptorLength = \
     construct.Rebuild(construct.Int8ul, construct.len_(construct.this)) \
     * "Descriptor Length"



#
# This file is part of usb-protocol.
#
"""
    Descriptors for USB Audio Class Devices (UAC), Release 3

    [Audio30] refers to "Universal Serial Bus Device Class Definition for Audio Devices", Release 3.0, September 22, 2016
    [Frmts30] refers to "Universal Serial Bus Device Class Definition for Audio Data Formats", Release 3.0, September 22, 2016
    [TermT30] refers to "Universal Serial Bus Device Class Definition for Terminal Types", Release 3.0, May 31, 2006

    NOTE: This is not complete yet and will be extended as needed
"""

from build.lib.usb_protocol.emitters import descriptor
import unittest
from enum import IntEnum

import construct
from   construct  import this, Default

from .. import LanguageIDs
from .standard import StandardDescriptorNumbers
from ..descriptor import \
    DescriptorField, DescriptorNumber, DescriptorFormat, \
    BCDFieldAdapter, DescriptorLength


class AudioInterfaceClassCode(IntEnum):
    # As defined in [Audio30], Table A-4
    AUDIO = 0x01


class AudioFunctionClassCode(IntEnum):
    # As defined in [Audio30], Table A-1
    AUDIO_FUNCTION = AudioInterfaceClassCode.AUDIO


class AudioFunctionSubclassCodes(IntEnum):
    # As defined in [Audio30], Table A-2
    FUNCTION_SUBCLASS_UNDEFINED = 0x00
    FULL_ADC_3_0                = 0x01
    GENERIC_IO                  = 0x20
    HEADPHONE                   = 0x21
    SPEAKER                     = 0x22
    MICROPHONE                  = 0x23
    HEADSET                     = 0x24
    HEADSET_ADAPTER             = 0x25
    SPEAKERPHONE                = 0x26


class AudioFunctionProtocolCodes(IntEnum):
    # As defined in [Audio30], Table A-3
    FUNCTION_PROTOCOL_UNDEFINED = 0x00
    AF_VERSION_01_00 = AudioInterfaceProtocolCodes.IP_VERSION_01_00
    AF_VERSION_02_00 = AudioInterfaceProtocolCodes.IP_VERSION_02_00
    AF_VERSION_03_00 = AudioInterfaceProtocolCodes.IP_VERSION_03_00


class AudioInterfaceSubclassCodes(IntEnum):
    # As defined in [Audio30], Table A-5
    INTERFACE_SUBCLASS_UNDEFINED = 0x00
    AUDIO_CONTROL                = 0x01
    AUDIO_STREAMING              = 0x02
    MIDI_STREAMING               = 0x03


class AudioInterfaceProtocolCodes(IntEnum):
    # As defined in [Audio30], Table A-6
    IP_VERSION_01_00 = 0x00
    IP_VERSION_02_00 = 0x20
    IP_VERSION_03_00 = 0x30


class AudioFunctionCategoryCodes(IntEnum):
    # As defined in [Audio30], Table A-7
    FUNCTION_SUBCLASS_UNDEFINED = 0x00
    DESKTOP_SPEAKER             = 0x01
    HOME_THEATER                = 0x02
    MICROPHONE                  = 0x03
    HEADSET                     = 0x04
    TELEPHONE                   = 0x05
    CONVERTER                   = 0x06
    VOICE_SOUND_RECORDER        = 0x07
    IO_BOX                      = 0x08
    MUSICAL_INSTRUMENT          = 0x09
    PRO_AUDIO                   = 0x0A
    AUDIO_VIDEO                 = 0x0B
    CONTROL_PANEL               = 0x0C
    HEADPHONE                   = 0x0D
    GENERIC_SPEAKER             = 0x0E
    HEADSET_ADAPTER             = 0x0F
    SPEAKERPHONE                = 0x10
    OTHER                       = 0xFF


class AudioClassSpecificDescriptorTypes(IntEnum):
    # As defined in [Audio30], Table A-8
    CS_UNDEFINED     = 0x20
    CS_DEVICE        = 0x21
    CS_CONFIGURATION = 0x22
    CS_STRING        = 0x23
    CS_INTERFACE     = 0x24
    CS_ENDPOINT      = 0x25
    CS_CLUSTER       = 0x26


class ClusterDescriptorSubtypes(IntEnum):
    # As defined in [Audio30], Table A-9
    SUBTYPE_UNDEFINED = 0x00


class ClusterDescriptorSegmentTypes(IntEnum):
    # As defined in [Audio30], Table A-10
    SEGMENT_UNDEFINED      = 0x00
    CLUSTER_DESCRIPTION    = 0x01
    CLUSTER_VENDOR_DEFINED = 0x1F
    CHANNEL_INFORMATION    = 0x20
    CHANNEL_AMBISONIC      = 0x21
    CHANNEL_DESCRIPTION    = 0x22
    CHANNEL_VENDOR_DEFINED = 0xFE
    END_SEGMENT            = 0xFF


class  ChannelPurposeDefinitions(IntEnum):
    # As defined in [Audio30], Table A-11
    PURPOSE_UNDEFINED = 0x00
    GENERIC_AUDIO     = 0x01
    VOICE             = 0x02
    SPEECH            = 0x03
    AMBIENT           = 0x04
    REFERENCE         = 0x05
    ULTRASONIC        = 0x06
    VIBROKINETIC      = 0x07
    NON_AUDIO         = 0xFF


class AmbisonicComponentOrderingConventionTypes(IntEnum):
    # As defined in [Audio30], Table A-13
    ORD_TYPE_UNDEFINED           = 0x00
    AMBISONIC_CHANNEL_NUMBER_ACN = 0x01
    FURSE_MALHAM                 = 0x02
    SINGLE_INDEX_DESIGNATION_SID = 0x03


class AmbisonicNormalizationTypes(IntEnum):
    # As defined in [Audio30], Table A-14
    NORM_TYPE_UNDEFINED = 0x00
    MAX_N               = 0x01
    SN3D                = 0x02
    N3D                 = 0x03
    SN2D                = 0x04
    N2D                 = 0x05


class AudioClassSpecificACInterfaceDescriptorSubtypes(IntEnum):
    # As defined in [Audio30], Table A-15
    AC_DESCRIPTOR_UNDEFINED  = 0x00
    HEADER                   = 0x01
    INPUT_TERMINAL           = 0x02
    OUTPUT_TERMINAL          = 0x03
    EXTENDED_TERMINAL        = 0x04
    MIXER_UNIT               = 0x05
    SELECTOR_UNIT            = 0x06
    FEATURE_UNIT             = 0x07
    EFFECT_UNIT              = 0x08
    PROCESSING_UNIT          = 0x09
    EXTENSION_UNIT           = 0x0A
    CLOCK_SOURCE             = 0x0B
    CLOCK_SELECTOR           = 0x0C
    CLOCK_MULTIPLIER         = 0x0D
    SAMPLE_RATE_CONVERTER    = 0x0E
    CONNECTORS               = 0x0F
    POWER_DOMAIN             = 0x10


class AudioClassSpecificASInterfaceDescriptorSubtypes(IntEnum):
    # As defined in [Audio30], Table A-16
    AS_DESCRIPTOR_UNDEFINED = 0x00
    AS_GENERAL              = 0x01
    AS_VALID_FREQ_RANGE     = 0x02


class AudioClassSpecificStringDescriptorSubtypes(IntEnum):
    # As defined in [Audio30], Table A-17
    SUBTYPE_UNDEFINED = 0x00


class ExtendedTerminalSegmentTypes(IntEnum):
    # As defined in [Audio30], Table A-18
    SEGMENT_UNDEFINED                = 0x00
    TERMINAL_VENDOR_DEFINED          = 0x1F
    CHANNEL_BANDWIDTH                = 0x20
    CHANNEL_MAGNITUDE_RESPONSE       = 0x21
    CHANNEL_MAGNITUDE_PHASE_RESPONSE = 0x22
    CHANNEL_POSITION_XYZ             = 0x23
    CHANNEL_POSITION_R_THETA_PHI     = 0x24
    CHANNEL_VENDOR_DEFINED           = 0xFE
    END_SEGMENT                      = 0xFF


class EffectUnitEffectTypes(IntEnum):
    # As defined in [Audio30], Table A-19
    EFFECT_UNDEFINED        = 0x00
    PARAM_EQ_SECTION_EFFECT = 0x01
    REVERBERATION_EFFECT    = 0x02
    MOD_DELAY_EFFECT        = 0x03
    DYN_RANGE_COMP_EFFECT   = 0x04


class ProcessingUnitProcessTypes(IntEnum):
    # As defined in [Audio30], Table A20
    PROCESS_UNDEFINED       = 0x0000
    UP_DOWNMIX_PROCESS      = 0x0001
    STEREO_EXTENDER_PROCESS = 0x0002
    MULTI_FUNCTION_PROCESS  = 0x0003


class AudioClassSpecificEndpointDescriptorSubtypes(IntEnum):
    # As defined in [Audio30], Table A-21
    DESCRIPTOR_UNDEFINED = 0x00
    EP_GENERAL           = 0x01


class AudioClassSpecificRequestCodes(IntEnum):
    # As defined in [Audio30], Table A-22
    REQUEST_CODE_UNDEFINED     = 0x00
    CUR                        = 0x01
    RANGE                      = 0x02
    MEM                        = 0x03
    INTEN                      = 0x04
    STRING                     = 0x05
    HIGH_CAPABILITY_DESCRIPTOR = 0x06


class AudioControlInterfaceControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-23
    AC_CONTROL_UNDEFINED        = 0x00
    AC_ACTIVE_INTERFACE_CONTROL = 0x01
    AC_POWER_DOMAIN_CONTROL     = 0x02


class ClockSourceControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-24
    CS_CONTROL_UNDEFINED   = 0x00
    CS_SAM_FREQ_CONTROL    = 0x01
    CS_CLOCK_VALID_CONTROL = 0x02


class ClockSelectorControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-24
    CX_CONTROL_UNDEFINED      = 0x00
    CX_CLOCK_SELECTOR_CONTROL = 0x01


class ClockMultiplierControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-26
    CM_CONTROL_UNDEFINED   = 0x00
    CM_NUMERATOR_CONTROL   = 0x01
    CM_DENOMINATOR_CONTROL = 0x02


class TerminalControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-27
    TE_CONTROL_UNDEFINED = 0x00
    TE_INSERTION_CONTROL = 0x01
    TE_OVERLOAD_CONTROL  = 0x02
    TE_UNDERFLOW_CONTROL = 0x03
    TE_OVERFLOW_CONTROL  = 0x04
    TE_LATENCY_CONTROL   = 0x05


class MixerControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-28
    MU_CONTROL_UNDEFINED = 0x00
    MU_MIXER_CONTROL     = 0x01
    MU_UNDERFLOW_CONTROL = 0x02
    MU_OVERFLOW_CONTROL  = 0x03
    MU_LATENCY_CONTROL   = 0x04


class SelectorControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-29
    SU_CONTROL_UNDEFINED = 0x00
    SU_SELECTOR_CONTROL  = 0x01
    SU_LATENCY_CONTROL   = 0x02


class FeatureUnitControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-30
    FU_CONTROL_UNDEFINED         = 0x00
    FU_MUTE_CONTROL              = 0x01
    FU_VOLUME_CONTROL            = 0x02
    FU_BASS_CONTROL              = 0x03
    FU_MID_CONTROL               = 0x04
    FU_TREBLE_CONTROL            = 0x05
    FU_GRAPHIC_EQUALIZER_CONTROL = 0x06
    FU_AUTOMATIC_GAIN_CONTROL    = 0x07
    FU_DELAY_CONTROL             = 0x08
    FU_BASS_BOOST_CONTROL        = 0x09
    FU_LOUDNESS_CONTROL          = 0x0A
    FU_INPUT_GAIN_CONTROL        = 0x0B
    FU_INPUT_GAIN_PAD_CONTROL    = 0x0C
    FU_PHASE_INVERTER_CONTROL    = 0x0D
    FU_UNDERFLOW_CONTROL         = 0x0E
    FU_OVERFLOW_CONTROL          = 0x0F
    FU_LATENCY_CONTROL           = 0x10


class ParametricEqualizerSectionEffectUnitControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-31
    PE_CONTROL_UNDEFINED  = 0x00
    PE_ENABLE_CONTROL     = 0x01
    PE_CENTERFREQ_CONTROL = 0x02
    PE_QFACTOR_CONTROL    = 0x03
    PE_GAIN_CONTROL       = 0x04
    PE_UNDERFLOW_CONTROL  = 0x05
    PE_OVERFLOW_CONTROL   = 0x06
    PE_LATENCY_CONTROL    = 0x07


class ReverberationEffectUnitControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-32
    RV_CONTROL_UNDEFINED      = 0x00
    RV_ENABLE_CONTROL         = 0x01
    RV_TYPE_CONTROL           = 0x02
    RV_LEVEL_CONTROL          = 0x03
    RV_TIME_CONTROL           = 0x04
    RV_FEEDBACK_CONTROL       = 0x05
    RV_PREDELAY_CONTROL       = 0x06
    RV_DENSITY_CONTROL        = 0x07
    RV_HIFREQ_ROLLOFF_CONTROL = 0x08
    RV_UNDERFLOW_CONTROL      = 0x09
    RV_OVERFLOW_CONTROL       = 0x0A
    RV_LATENCY_CONTROL        = 0x0B


class ModulationDelayEffectUnitControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-33
    MD_CONTROL_UNDEFINED = 0x00
    MD_ENABLE_CONTROL    = 0x01
    MD_BALANCE_CONTROL   = 0x02
    MD_RATE_CONTROL      = 0x03
    MD_DEPTH_CONTROL     = 0x04
    MD_TIME_CONTROL      = 0x05
    MD_FEEDBACK_CONTROL  = 0x06
    MD_UNDERFLOW_CONTROL = 0x07
    MD_OVERFLOW_CONTROL  = 0x08
    MD_LATENCY_CONTROL   = 0x09


class DynamicRangeCompressorEffectUnitControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-34
    DR_CONTROL_UNDEFINED        = 0x00
    DR_ENABLE_CONTROL           = 0x01
    DR_COMPRESSION_RATE_CONTROL = 0x02
    DR_MAXAMPL_CONTROL          = 0x03
    DR_THRESHOLD_CONTROL        = 0x04
    DR_ATTACK_TIME_CONTROL      = 0x05
    DR_RELEASE_TIME_CONTROL     = 0x06
    DR_UNDERFLOW_CONTROL        = 0x07
    DR_OVERFLOW_CONTROL         = 0x08
    DR_LATENCY_CONTROL          = 0x09


class UpDownMixProcessingUnitControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-35
    UD_CONTROL_UNDEFINED   = 0x00
    UD_MODE_SELECT_CONTROL = 0x01
    UD_UNDERFLOW_CONTROL   = 0x02
    UD_OVERFLOW_CONTROL    = 0x03
    UD_LATENCY_CONTROL     = 0x04


class StereoExtenderProcessingUnitControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-36
    ST_EXT_CONTROL_UNDEFINED = 0x00
    ST_EXT_WIDTH_CONTROL     = 0x01
    ST_EXT_UNDERFLOW_CONTROL = 0x02
    ST_EXT_OVERFLOW_CONTROL  = 0x03
    ST_EXT_LATENCY_CONTROL   = 0x04


class ExtensionUnitControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-37
    XU_CONTROL_UNDEFINED = 0x00
    XU_UNDERFLOW_CONTROL = 0x01
    XU_OVERFLOW_CONTROL  = 0x02
    XU_LATENCY_CONTROL   = 0x03


class AudioStreamingInterfaceControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-38
    AS_CONTROL_UNDEFINED         = 0x00
    AS_ACT_ALT_SETTING_CONTROL   = 0x01
    AS_VAL_ALT_SETTINGS_CONTROL  = 0x02
    AS_AUDIO_DATA_FORMAT_CONTROL = 0x03


class EndpointControlSelectors(IntEnum):
    # As defined in [Audio30], Table A-39
    EP_CONTROL_UNDEFINED     = 0x00
    EP_PITCH_CONTROL         = 0x01
    EP_DATA_OVERRUN_CONTROL  = 0x02
    EP_DATA_UNDERRUN_CONTROL = 0x03


class USBTerminalTypes(IntEnum):
    # As defined in [TermT30], Table 2-1
    USB_UNDEFINED       = 0x0100
    USB_STREAMING       = 0x0101
    USB_VENDOR_SPECIFIC = 0x01FF


class InputTerminalTypes(IntEnum):
    # As defined in [TermT30], Table 2-2
    INPUT_UNDEFINED             = 0x0200
    MICROPHONE                  = 0x0201
    DESKTOP_MICROPHONE          = 0x0202
    PERSONAL_MICROPHONE         = 0x0203
    OMNI_DIRECTIONAL_MICROPHONE = 0x0204
    MICROPHONE_ARRAY            = 0x0205
    PROCESSING_MICROPHONE_ARRAY = 0x0206


class OutputTerminalTypes(IntEnum):
    # As defined in [TermT30], Table 2-3
    OUTPUT_UNDEFINED              = 0x0300
    SPEAKER                       = 0x0301
    HEADPHONES                    = 0x0302
    DESKTOP_SPEAKER               = 0x0304
    ROOM_SPEAKER                  = 0x0305
    COMMUNICATION_SPEAKER         = 0x0306
    LOW_FREQUENCY_EFFECTS_SPEAKER = 0x0307


class BidirectionalTerminalTypes(IntEnum):
    # As defined in [TermT30], Table 2-4
    BIDIRECTIONAL_UNDEFINED       = 0x0400
    HANDSET                       = 0x0401
    HEADSET                       = 0x0402
    ECHO_SUPPRESSING_SPEAKERPHONE = 0x0404
    ECHO_CANCELING_SPEAKERPHONE   = 0x0405


class TelephonyTerminalTypes(IntEnum):
    # As defined in [TermT30], Table 2-5
    TELEPHONY_UNDEFINED = 0x0500
    PHONE_LINE          = 0x0501
    TELEPHONE           = 0x0502
    DOWN_LINE_PHONE     = 0x0503


class ExternalTerminalTypes(IntEnum):
    # As defined in [TermT30], Table 2-6
    EXTERNAL_UNDEFINED             = 0x0600
    ANALOG_CONNECTOR               = 0x0601
    DIGITAL_AUDIO_INTERFACE        = 0x0602
    LINE_CONNECTOR                 = 0x0603
    SPDIF_INTERFACE                = 0x0605
    IEEE_1394_DA_STREAM            = 0x0606
    IEEE_1394_DV_STREAM_SOUNDTRACK = 0x0607
    ADAT_LIGHTPIPE                 = 0x0608
    TDIF                           = 0x0609
    MADI                           = 0x060A


class EmbeddedFunctionTerminalTypes(IntEnum):
    # As defined in [TermT30], Table 2-7
    EMBEDDED_UNDEFINED       = 0x0700
    EQUALIZATION_NOISE       = 0x0702
    CD_PLAYER                = 0x0703
    DAT                      = 0x0704
    DCC                      = 0x0705
    ANALOG_TAPE              = 0x0707
    PHONOGRAPH               = 0x0708
    VCR_AUDIO                = 0x0709
    VIDEO_DISC_AUDIO         = 0x070A
    DVD_AUDIO                = 0x070B
    TV_TUNER_AUDIO           = 0x070C
    SATELLITE_RECEIVER_AUDIO = 0x070D
    CABLE_TUNER_AUDIO        = 0x070E
    DSS_AUDIO                = 0x070F
    RADIO_RECEIVER           = 0x0710
    RADIO_TRANSMITTER        = 0x0711
    MULTI_TRACK_RECORDER     = 0x0712
    SYNTHESIZER              = 0x0713
    PIANO                    = 0x0714
    GUITAR                   = 0x0715
    DRUMS_RHYTHM             = 0x0716
    OTHER_MUSICAL_INSTRUMENT = 0x0717


class ConnectorTypes(IntEnum):
    UNDEFINED                     = 0x00
    PHONE_CONNECTOR_2_5_MM        = 0x01
    PHONE_CONNECTOR_3_5_MM        = 0x02
    PHONE_CONNECTOR_6_35_MM       = 0x03
    XLR_6_35MM_COMBO_CONNECTOR    = 0x04
    XLR                           = 0x05
    OPTICAL_3_5MM_COMBO_CONNECTOR = 0x06
    RCA                           = 0x07
    BNC                           = 0x08
    BANANA                        = 0x09
    BINDING_POST                  = 0x0A
    SPEAKON                       = 0x0B
    SPRING_CLIP                   = 0x0C
    SCREW_TYPE                    = 0x0D
    DIN                           = 0x0E
    MINI_DIN                      = 0x0F
    EUROBLOCK                     = 0x10
    USB_TYPE_C                    = 0x11
    RJ_11                         = 0x12
    RJ_45                         = 0x13
    TOSLINK                       = 0x14
    HDMI                          = 0x15
    Mini_HDMI                     = 0x16
    Micro_HDMI                    = 0x17
    DP                            = 0x18
    MINI_DP                       = 0x19
    D_SUB                         = 0x1A
    THUNDERBOLT                   = 0x1B
    LIGHTNING                     = 0x1C
    WIRELESS                      = 0x1D
    USB_STANDARD_A                = 0x1E
    USB_STANDARD_B                = 0x1F
    USB_MINI_B                    = 0x20
    USB_MICRO_B                   = 0x21
    USB_MICRO_AB                  = 0x22
    USB_3_0_MICRO_B               = 0x23

class AudioDataFormats(IntEnum):
    PCM                             = (1 << 0)
    PCM8                            = (1 << 1)
    IEEE_FLOAT                      = (1 << 2)
    ALAW                            = (1 << 3)
    MULAW                           = (1 << 4)
    DSD                             = (1 << 5)
    RAW_DATA                        = (1 << 6)
    PCM_IEC60958                    = (1 << 7)
    AC_3                            = (1 << 8)
    MPEG_1_Layer1                   = (1 << 9)
    MPEG_1_Layer2_3                 = (1 << 10) # These share the same bit
    MPEG_2_NOEXT                    = (1 << 10) # These share the same bit
    MPEG_2_EXT                      = (1 << 11)
    MPEG_2_AAC_ADTS                 = (1 << 12)
    MPEG_2_Layer1_LS                = (1 << 13)
    MPEG_2_Layer2_3_LS              = (1 << 14)
    DTS_I                           = (1 << 15)
    DTS_II                          = (1 << 16)
    DTS_III                         = (1 << 17)
    ATRAC                           = (1 << 18)
    ATRAC2_3                        = (1 << 19)
    WMA                             = (1 << 20)
    E_AC_3                          = (1 << 21)
    MAT                             = (1 << 22)
    DTS_IV                          = (1 << 23)
    MPEG_4_HE_AAC                   = (1 << 24)
    MPEG_4_HE_AAC_V2                = (1 << 25)
    MPEG_4_AAC_LC                   = (1 << 26)
    DRA                             = (1 << 27)
    MPEG_4_HE_AAC_SURROUND          = (1 << 28)
    MPEG_4_AAC_LC_SURROUND          = (1 << 29)
    MPEG_H_3D_AUDIO                 = (1 << 30)
    AC4                             = (1 << 31)
    MPEG_4_AAC_ELD                  = (1 << 32)


# As defined in [Audio30], Table 4-47
AudioControlInterruptEndpointDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(7, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(AudioClassSpecificStandardDescriptorNumbers.CS_ENDPOINT),
    "bEndpointAddress"    / DescriptorField(description="The address of the endpoint, use USBDirection.*.from_endpoint_address()"),
    "bmAttributes"        / DescriptorField(description="D1..0: Transfer type (0b11 = Interrupt)", default=0b11),
    "wMaxPacketSize"      / DescriptorField(description="Maximum packet size this endpoint is capable of. Used here to pass 6-byte interrupt information.", default=6),
    "bInterval"           / DescriptorField(description="Interval for polling the Interrupt endpoint")
)

# As defined in [Audio30], Table 4-51
AudioStreamingIsochronousEndpointDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(7, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(StandardDescriptorNumbers.ENDPOINT),
    "bEndpointAddress"    / DescriptorField(description="The address of the endpoint, use USBDirection.*.from_endpoint_address()"),
    "bmAttributes"        / DescriptorField(description="D1..0: transfer type (01=isochronous); D3..2: synchronization type (01=asynchronous/10=adaptive/11=synchronous); D5..4: usage (00=data/10=feedback)", default=0b000101),
    "wMaxPacketSize"      / DescriptorField(description="Maximum packet size this endpoint is capable of. Used here to pass 6-byte interrupt information.", default=6),
    "bInterval"           / DescriptorField(description="Interval for polling the Interrupt endpoint")
)

# As defined in [Audio30], Table 4-53
AudioStreamingIsochronousFeedbackEndpointDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(7, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(StandardDescriptorNumbers.ENDPOINT),
    "bEndpointAddress"    / DescriptorField(description="The address of the endpoint, use USBDirection.*.from_endpoint_address()"),
    "bmAttributes"        / DescriptorField(description="D1..0: transfer type (01=isochronous); D3..2: synchronization type (00=no sync); D5..4: usage (10=feedback)", default=0b00100001),
    "wMaxPacketSize"      / DescriptorField(description="Maximum packet size this endpoint is capable of. Used here to pass 6-byte interrupt information.", default=6),
    "bInterval"           / DescriptorField(description="Interval for polling the Interrupt endpoint")
)

HeaderDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(10, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(AudioClassSpecificStandardDescriptorNumbers.CS_INTERFACE),
    "bDescriptorSubtype"  / DescriptorNumber(AudioClassSpecificACInterfaceDescriptorSubtypes.HEADER),
    "bCategory"           / DescriptorField(description="Audio Function Category, see AudioFunctionCategoryCodes"),
    "wTotalLength"        / DescriptorField("Length including subordinates"),
    "bmControls"          / DescriptorField("D1..0: Latency Control; D31..2: Reserved.", length=4, default=0)
)

AudioStreamingInterfaceDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(9, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(DescriptorTypes.INTERFACE),
    "bInterfaceNumber"    / DescriptorField(description="ID of the streaming interface"),
    "bAlternateSetting"   / DescriptorField(description="alternate setting number for the interface", default=0),
    "bNumEndpoints"       / DescriptorField(description="Number of data endpoints used (excluding endpoint 0). Can be: 0 (no data endpoint); 1 (data endpoint); 2 (data + explicit feedback endpoint)", default=0),
    "bInterfaceClass"     / DescriptorNumber(AudioInterfaceClassCode.AUDIO),
    "bInterfaceSubClass"  / DescriptorNumber(AudioInterfaceSubclassCodes.AUDIO_STREAMING),
    "bInterfaceProtocol"  / DescriptorNumber(AudioInterfaceProtocolCodes.IP_VERSION_03_00),
    "iInterface"          / DescriptorField(description="index of a string descriptor describing this interface (0 = unused)")
)

ClassSpecificAudioStreamingInterfaceDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(23, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(AudioClassSpecificStandardDescriptorNumbers.CS_INTERFACE),
    "bDescriptorSubtype"  / DescriptorNumber(AudioClassSpecificASInterfaceDescriptorSubtypes.AS_GENERAL),
    "bTerminalLink"       / DescriptorField(description="the ID of the terminal to which this interface is connected"),
    "bmControls"          / DescriptorField(description="D1..0: active alternate setting control; D3..2: valid alternate settings control; D5..4: audio data format control; D31..6: reserved"),
    "wClusterDescrID"     / DescriptorField(description="ID of the cluster descriptor of the audio streamin interface"),
    "bmFormats"           / DescriptorField(description="audio data formats which can be used with this interface", length=8, default=AudioDataFormats.PCM),
    "bSubslotSize"        / DescriptorField(description="number of bytes occupied by one audio subslot"),
    "bBitResolution"      / DescriptorField(description="number of effectively used bits in the audio subslot"),
    "bmAuxProtocols"      / DescriptorField(description="which auxiliary protocols are required", default=0),
    "bControlSize"        / DescriptorField(description="size of the control channel words in bytes")
)

InputTerminalDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(20, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(AudioClassSpecificStandardDescriptorNumbers.CS_INTERFACE),
    "bDescriptorSubtype"  / DescriptorNumber(AudioClassSpecificACInterfaceDescriptorSubtypes.INPUT_TERMINAL),
    "bTerminalID"         / DescriptorField(description="unique identifier for the terminal within the audio function."),
    "wTerminalType"       / DescriptorField(description="a value of one of the terminal types Enums (eg InputTerminaTypes, ExternalTerminalTypes)"),
    "bAssocTerminal"      / DescriptorField(description="ID of the associated output terminal"),
    "bCSourceID"          / DescriptorField(description="ID of the clock which is connected to this terminal"),
    "bmControls"          / DescriptorField(description="D1..0: Insertion Control; D3..2: Overload Control; D5..4: Underflow Control; D7..6: Overflow Control; D31..8: Reserved"),
    "wClusterDescrID"     / DescriptorField(description="ID of the cluster descriptor for this input terminal."),
    "wExTerminalDescrID"  / DescriptorField(description="ID of the extended terminal descriptor for this input terminal. Zero if no extended terminal descriptor is present."),
    "wConnectorsDescrID"  / DescriptorField(description="ID of the Connectors descriptor for this Input Terminal. Zero if no connectors descriptor is present."),
    "wTerminalDescrStr"   / DescriptorField(description="ID of a class-specific string descriptor, describing the input terminal.")
)

OutputTerminalDescriptor = DescriptorFormat(
    "bLength"             / construct.Const(19, construct.Int8ul),
    "bDescriptorType"     / DescriptorNumber(AudioClassSpecificStandardDescriptorNumbers.CS_INTERFACE),
    "bDescriptorSubtype"  / DescriptorNumber(AudioClassSpecificACInterfaceDescriptorSubtypes.OUTPUT_TERMINAL),
    "bTerminalID"         / DescriptorField(description="unique identifier for the terminal within the audio function."),
    "wTerminalType"       / DescriptorField(description="a value of one of the terminal types Enums (eg OutputTerminaTypes, ExternalTerminalTypes)"),
    "bAssocTerminal"      / DescriptorField(description="ID of the associated input terminal"),
    "bSourceID"           / DescriptorField(description="ID of the unit or terminal which is connected to this terminal"),
    "bCSourceID"          / DescriptorField(description="ID of the clock which is connected to this terminal"),
    "bmControls"          / DescriptorField(description="D1..0: Insertion Control; D3..2: Overload Control; D5..4: Underflow Control; D7..6: Overflow Control; D31..8: Reserved"),
    "wClusterDescrID"     / DescriptorField(description="ID of the cluster descriptor for this input terminal."),
    "wExTerminalDescrID"  / DescriptorField(description="ID of the extended terminal descriptor for this output terminal. Zero if no extended terminal descriptor is present.", default=0),
    "wConnectorsDescrID"  / DescriptorField(description="ID of the connectors descriptor for this input terminal. Zero if no connectors descriptor is present.", default=0),
    "wTerminalDescrStr"   / DescriptorField(description="ID of a class-specific string descriptor, describing the output terminal.")
)



from setuptools import setup, find_packages

setup(

    # Vitals
    name='usb_descriptors',
    license='BSD',
    url='https://github.com/hansfbaier/python-usb-descriptors',
    author='Hans Baier',
    author_email='hansfbaier@gmail.com',
    description='python library providing utilities, data structures, constants, parsers, and tools for working with USB data',
    use_scm_version= {
        "root": '..',
        "relative_to": __file__,
        "version_scheme": "guess-next-dev",
        "local_scheme": lambda version : version.format_choice("+{node}", "+{node}.dirty"),
        "fallback_version": "0.0"
    },

    # Imports / exports / requirements.
    platforms='any',
    packages=find_packages(),
    include_package_data=True,
    python_requires="~=3.7",
    install_requires=['construct'],
    setup_requires=['setuptools', 'setuptools_scm', 'wheel'],

    # Metadata
    classifiers = [
        'Programming Language :: Python',
        'Development Status :: 1 - Planning',
        'Natural Language :: English',
        'Environment :: Console',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering',
        'Topic :: Security',
        ],
)



#
# This file is part of usb-protocol.
#
""" USB-related emitters. """

from .construct_interop    import emitter_for_format, ConstructEmitter
from .descriptors.standard import DeviceDescriptorCollection, SuperSpeedDeviceDescriptorCollection
